@echo off

rem	Copyright (c) 2016 Dave Cannon.  All rights reserved.
rem	$Id: bcpl.cmd,v 3.00 30-Mar-2016 20:27:27 rn Exp $

setlocal enabledelayedexpansion
if errorlevel 1 (
  echo Unable to enable delayed expansion - aborting!
  goto eof
)

set vn=3
set rvn=00
set edt=30-Mar-2016

set version=%0 Vn %vn%.%rvn%	%edt%

rem	Vn 3.00	30-Mar-2016 amended for BCPL 1981-03-05 (IBM 360/370) version
rem	Vn 2.03	20-Aug-2014 all environment variables now set externally by ..\BCPLenv.cmd
rem	Vn 2.02	24-Aug-2013 introduction of 'getopt' to parse command line options
rem	Vn 2.01	20-Aug-2013	Add variable for MinGW version number...
rem	Vn 2.00	13-Jul-2013	Initial version for DOS under Windows

if not '%1'=='' goto doit

:usage
echo	Usage:	%0 [-Options] [-o output] file
echo		where	file is a BCPL source program to be compiled
echo			[-o output] (optional) specifies the compiled program name
echo			[-Options] may be any of:
echo			[-O] (optional) suppresses the code optimiser
echo			[-c] (optional) preserve OCODE output as 'file.ocd'
echo			[-i] (optional) preserve INTCODE output as 'file.i'
echo			[-l] (optional) preserve the 'file.o' object file for ld
echo			[-S] (optional) preserve the 'file.asm' assembler file
echo.
echo		%version%
echo.
goto eof

:doit

set oflag=0
set args=

if not defined GnuWin (
  echo.
  echo 	The BCPL environment variables have not been established.
  echo 	You must execute ..\BCPLenv.cmd before running this script
  echo.
  goto eof
)

set LDFLAGS=-Bdynamic -T"%GnuWin%\mingw32\lib\ldscripts\i386pe_mgw.x" "%GnuWin%\lib\crt2.o" "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtbegin.o"
set LDLIBS=-lmingw32 -lmoldname -lmingwex -lmsvcrt -luser32 -lkernel32 -ladvapi32 -lshell32 "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtend.o"
set BCPLcmpl=%BCPLroot%\BCPL 1981-03-05
set BCPLvn=%BCPLroot%\BCPL 1979-09-01\codegens

for /f "usebackq tokens=*" %%i in (`getopt cilOSo: %*`) do set args=%%i
for %%i in (%args%) do (
	if '!oarg!'=='?' (
	  set oarg=%%i
	)
	if '!farg!'=='?' (
	  set farg=%%i
	)
	if '%%i'=='-O' (
	  set oflag=1
	)
	if '%%i'=='-o' (
	  set oarg=?
	)
	if '%%i'=='-c' (
	  set ocd=1
	)
	if '%%i'=='-i' (
	  set int=1
	)
	if '%%i'=='-l' (
	  set obj=1
	)
	if '%%i'=='-S' (
	  set asm=1
	)
	if '%%i'=='--' (
	  set farg=?
	)
)
if '%farg%'=='?' goto usage
if '%oarg%'=='?' goto usage

if not exist %farg% (
	echo "%0 %farg%: no such file"
	goto usage
	)

if not exist libhdr cp "%BCPLhdrs%\libhdr" libhdr

if defined oarg (
    set f=%oarg%
) else (
	for /f "usebackq" %%i in (`basename %farg% .b`) do set f=%%i
)

if exist %f%.asm del /q/f %f%.asm > nul
if exist %f%.o del /q/f %f%.o > nul
if not exist params cp "%BCPLcmpl%\params" .

"%BCPLcmpl%\bcpl+" %farg%
mv OCODE %f%.ocd 
mv CODE %f%360.asm > nul

if %oflag%==0 (
	"%BCPLvn%\cg-intcode" %f%.ocd %f%.i
)

if not defined ocd if exist %f%.ocd del /q/f %f%.ocd > nul

cat %f%.i "%BCPLlib%\blib.i" "%BCPLlib%\iclib.i" > %f%.int
"%BCPLlib%\xg_i" < %f%.int
if not defined int del /q/f %f%.i > nul
if exist %f%.int del /q/f %f%.int > nul

if %oflag% equ 1 (
	"%BCPLcopt%\copt" "%BCPLsrc%\rules" < ASM > ASM.op && mv ASM.op ASM
)
move ASM %f%.asm > nul
as -o %f%.o %f%.asm
if not defined obj ld %LDFLAGS% -o %f%.exe %f%.o "%BCPLsrc%\su.o" "%BCPLsrc%\rtc.o" %LDLIBS%

if not defined asm (
  if exist %f%.asm del /q/f %f%.asm > nul
  if exist %f%360.asm del /q/f %f%360.asm > nul
)
if not defined obj if exist %f%.o del /q/f %f%.o > nul
if exist SYSGO del /q/f SYSGO > nul

:eof

set f=
set vn=
set rvn=
set edt=
set int=
set ocd=
set version=
set oflag=
set asm=
set obj=
set oname=
set args=
set LDFLAGS=
set LDLIBS=
