// Copyright (c) 2004, 2012 Robert Nordier.  All rights reserved.

// BCPL compiler backend: generates x86 assembler from INTCODE
// Amended to accept INTCODE from Martin Richards' current BCPL compiler
//				 									2013 Dave Cannon

// $Id: xg.b,v 4.03 2013/11/10 16:14:51 rn Exp $

GET "LIBHDR"

MANIFEST $( WORDSZ=4; GSZ=1000; LSZ=10000		// GSZ originally 500, LSZ originally 500
			LINC=10000 $)						// LINC originally 500				DC: 29-March-2013

// Type values (N=number, P=local, G=global, L=static, R=register)
// at up to two levels of indirection
MANIFEST $(
T.I =#X04										// The indirect bit, set at *** below (which may become...
T.II=#X08										// ...for double indirection)

T.N =#X00
T.LP=#X01
T.LG=#X02
T.LL=#X03

T.R= #X04										// = T.N  | T.I
T.P= #X05										// = T.LP | T.I
T.G= #X06										// = T.LG | T.I
T.L= #X07										// = T.LL | T.I

T.IR=#X08										// = T.N  | T.II
T.IP=#X09										// = T.LP | T.II
T.IG=#X0A										// = T.LG | T.II
T.IL=#X0B										// = T.LL | T.II

T.LF=#X10										// T.LF added						DC: 6-August-2013
$)
MANIFEST $( UPB=11 $)							// Maximum SECTION string length	DC: 7-August-2013

// Assembler statement patterns for code generation: with the
// jump instructions first, as they need to be distinguished
MANIFEST $(
A.JE=0;   A.JNE=1;  A.JL=2;   A.JGE=3;  A.JG=4;   A.JLE=5;  A.JMP=6
A.MUL=7;  A.DIV=8;  A.MOV=9;  A.ADD=10; A.SUB=11; A.CMP=12; A.SHL=13
A.SHR=14; A.AND=15; A.OR=16;  A.XOR=17; A.MV2=18; A.LEA=19; A.SL2=20
A.SR2=21; A.SNE=22; A.SE=23;  A.SGE=24; A.SL=25;  A.SLE=26; A.SG=27
$)

GLOBAL $(
G:150           // Definitions vector for globals
L:151           // Definitions vector for labels
LN:152          // Labels in L vector
LOFF:153        // Label offset value
SECT:154        // Section (text or data) in output
LINE:155        // Source line number
COL:156         // Source column number
CH:157          // Source character last read
ERR:158         // Stream for errors
ASM:159         // Stream for output
S:160			// Container for SECTION name
RN:161			// Container for descriptive text
$)

LET START() = VALOF
$(  LET GVEC = GETVEC(GSZ)
    AND LVEC = GETVEC(LSZ)
	AND STR = VEC UPB/BYTESPERWORD
	AND RNAM = GETVEC(256/BYTESPERWORD)
	
    G := GVEC
    L := LVEC
	S := STR
	RN := RNAM
	PUTBYTE(RN, 0, 0)

    FOR I = 0 TO GSZ DO G!I := 0
    LN, SECT, LINE, COL := 0, 0, 0, 0
    LOFF := 20000				// Offset (originally 2000) to add to intcode label number - incremented by LINC on new segment of code
    CH := '*N'
    ERR := OUTPUT()
    ASM := FINDOUTPUT("ASM")
    SELECTOUTPUT(ASM)
    ASSEM()
    EPILOG()
    ENDWRITE()
    RESULTIS 0
$)

AND ASSEM() BE
$(  LET F, A, T, I = 0, 0, 0, 0
    RCH()
    F := CH
    SWITCHON F INTO $(
    DEFAULT:
        ERROR(1)
    CASE ENDSTREAMCH:
        GENER(0, 0, 0)
        RETURN
    CASE '*S': CASE '*N':
        LOOP
	CASE '$':									// Signals start of new routine, which may be labelled...	DC:	25-Oct-2013
		ENDCASE
	CASE 'N':									// SECTION label					DC: 7-August-2013
		RCH()
		A := RDN()								// String length
		IF A > UPB ERROR(20)
		FOR I=1 TO A DO $(
			RCH(); IF CH \= 'C' ERROR(20)
			RCH()
			PUTBYTE(S, I, RDN())
			$)
		PUTBYTE(S, 0, A)
		T := S
		ENDCASE
    CASE 'Z':
		PUTBYTE(RN, 0, 0)						// Clear any descriptive text
        LOFF := LOFF + LINC
        LOOP
    CASE 'G':
        RCH()
        A := RDN()
        UNLESS A<GSZ ERROR(2)
        UNLESS CH='L' ERROR(3)
        RCH()
        G!A := LOFF + RDN()
        LOOP
    CASE '0': CASE '1': CASE '2': CASE '3': CASE '4':
    CASE '5': CASE '6': CASE '7': CASE '8': CASE '9':
        F := 'B'
        A := LOFF + RDN()
        ENDCASE
    CASE 'A': CASE 'F': CASE 'J': CASE 'K':
    CASE 'L': CASE 'S': CASE 'T': CASE 'X':
        RCH()
        IF CH='I' $( I := 1; RCH() $)
        T := CH='G' -> T.LG,
             CH='P' -> T.LP,
             CH='L' -> T.LL,
			 CH='F' -> T.LF, T.N
        IF F='S' $( IF T=T.N ERROR(4); I := I + 1 $)
        IF T RCH()
        A := RDN()
        IF T=T.LL | T=T.LF A := A + LOFF
        T := I << 2 | T							// *** This makes some assumptions about the value of T and I!
        ENDCASE
    CASE 'C':
        RCH()
        A := RDN()
        ENDCASE
    CASE 'D':
        RCH()
        IF CH='L' $( T := T.LL; RCH() $)
        A := RDN()
        IF T=T.LL A := A + LOFF
    $)
    GENER(F, A, T)
$) REPEAT

AND GENER(F1, A1, T1) BE
$(  STATIC $(
		F=0;  A=0;  IA=0;  T=0
		F0=0; A0=0; IA0=0; T0=0
		F2=0; A2=0; IA2=0; T2=0
		F3=0; A3=0; IA3=0; T3=0
		F4=0; A4=0; IA4=0; T4=0
		F5=0; A5=0; IA5=0; T5=0
		F6=0; A6=0; IA6=0; T6=0
		F7=0; A7=0; IA7=0; T7=0
		F8=0; A8=0; IA8=0; T8=0
		F9=0; A9=0; IA9=0; T9=0
		F10=0; A10=0; IA10=0; T10=0
		XL=0
	$)

    LET XDYADIC(X) = 5<=X<=21 | X=32 | 35<=X<=37

    AND XFORCE(X, T) =
        (5<=X<=7 & T=T.N) |
        (16<=X<=17 & T NE T.N) |
        X>=32

    AND LOAD(R, J, FORCE) BE
    $(  LET E = T=T.LP | T=T.LG | T=T.LF
        AND M = E | (NOT J & (T=T.LL | T=T.LF))
    //	AND I = T>=T.IR
		AND I = T.IR<=T<=T.IL
        IF FORCE | M | I $(
            CODE1(E -> A.LEA, A.MOV, A, T, R)
            IF M & NOT J & NOT T=T.LF CODE1(A.SHR, 2, T.N, R)
            A, T := R, I -> T.IR, T.R
        $)
    $)

    LET X = ?
    IA := A							// Preserve initial value of A	DC: 21-Oct-2013

    SWITCHON F INTO $(
    CASE 0:
        ENDCASE
	CASE 'N':
		EMIT("#	SECTION @S", T)
		EMIT(".title @S", T)
        ENDCASE
	CASE '$':							// Entry point to new routine...
		UNLESS GETBYTE(RN, 0)=0 $(
			EMIT("#	@S", RN)			// Output descriptive text		DC: 22-Oct-2013
			PUTBYTE(RN, 0, 0)
		$)
		ENDCASE
	CASE 'B':
        UNLESS LN<LSZ ERROR(5)
        L!LN := A
        LN := LN + 1
        ENDCASE
    CASE 'D':
        SECT := 1
        DATA(".long @A", A, T)
        SECT := 0
        ENDCASE
    CASE 'C':
        SECT := 1
        DATA(".byte @A", A, 0)
        UNLESS F1='C' $(
			EMIT(".align 4,0")
		$)
        SECT := 0
        ENDCASE
    CASE 'L':
        X := F1='X' & XDYADIC(A1) -> 1, 0
        LOAD(X, FALSE, X=0 | XFORCE(A1, T))
        ENDCASE
    CASE 'A': CASE 'S':
        LOAD(1, FALSE, FALSE)
        CODE((F='A' -> A.ADD, A.MV2), A, T)
        ENDCASE
    CASE 'J':
        LOAD(1, TRUE, FALSE)
        CODE(A.JMP, A, T)
        ENDCASE
    CASE 'T': CASE 'F':
        UNLESS T=T.LL ERROR(6)
        X := A0
        UNLESS F0='X' & 10<=X<=15 $(
            EMIT("testl %eax,%eax")
            X := 11
        $)
        CODE((X - 10) NEQV (F='F' -> 1, 0), A, T)
        ENDCASE
    CASE 'K':
        UNLESS T=T.N $(
			SELECTOUTPUT(ERR)
			WRITEF("	F=%C, A=%N, T=%N*n	F1=%C, A1=%N, T1=%N*n", F, A, T, F1, A1, T1)
			ERROR(7)
			$)
        EMIT("movl %ebp,%ecx")
        EMIT("addl @A,%ebp", A << 2, T.N, 0, FALSE);
        EMIT("movl %ecx,(%ebp)")
        EMIT("movl $1f,4(%ebp)")
        EMIT("jmp **%eax")
        EMIT("1:")
        ENDCASE
    CASE 'X':
        SWITCHON A INTO $(
        DEFAULT: ERROR(8)
        CASE 1:												// (1)	A := Location (A)
            EMIT("movl (,%eax,4),%eax")
            ENDCASE
        CASE 2:												// (2)	A := -A
            EMIT("negl %eax")
            ENDCASE
        CASE 3:												// (3)	A := NOT A
            EMIT("xorl $-1,%eax")
            ENDCASE
        CASE 4:												// (4)	This causes a return from the current function or routine;
															//		by convention the result of a function is left in A.
															//		C := Location(P + 1)
															//		P := Location(P)
            EMIT("movl 4(%ebp),%ecx")
            EMIT("movl (%ebp),%ebp")
            EMIT("jmp **%ecx")
            ENDCASE
        CASE 5:												// (5)	A := B * A
            CODE(A.MUL, A0, T0)
            ENDCASE
        CASE 6: CASE 7:										// (6)	A := B / A
															// (7)	A := B REM A
            EMIT("cltd")
            CODE(A.DIV, A0, T0)
            IF A=7
                EMIT("movl %edx,%eax")
            ENDCASE
        CASE 8:												// (8)	A := B + A
            CODE(A.ADD, A0, T0)
            ENDCASE
        CASE 9:												// (9)	A := B - A
            CODE(A.SUB, A0, T0)
            ENDCASE
        CASE 10: CASE 11: CASE 12: CASE 13: CASE 14: CASE 15:	// (10)	A := B = A
															// (11)	A := B ~= A
															// (12)	A := B < A
															// (13)	A := B >= A
															// (14)	A := B > A
															// (15)	A := B <= A
            CODE(A.CMP, A0, T0)
            IF F1='F' | F1='T'
                ENDCASE
            EMIT(ASTR(A - 10 + A.SNE))
            EMIT("movzbl %al,%eax")
            EMIT("decl %eax")
            ENDCASE
        CASE 16:											// (16)	A := B LSHIFT A
        //	CODE(T0=T.N -> A.SHL, A.SL2, A0, T0)
        //	ENDCASE
            TEST T0=T.N & A0<32 DO							// Intel 80386 Ref Manual: "If a shift count greater than 31 is attempted, only the bottom five bits of the shift count are used."
                CODE(A.SHL, A0, T0)
            ELSE $(
				EMIT("movl @A,%ecx", A0, T0, 0, FALSE)
				EMIT("jcxz 1f")
				EMIT("2:")
				EMIT("decl %ecx")
				EMIT("shll $1,%eax")
				EMIT("cmpl $32,%ecx")						// shift value still ge 32?
				EMIT("jge 2b")
				EMIT("shll %cl,%eax")
				EMIT("1:")
            $)
            ENDCASE
        CASE 17:											// (17)	A := B RSHIFT A
        //	CODE(T0=T.N -> A.SHR, A.SR2, A0, T0)
        //	ENDCASE
            TEST T0=T.N & A0<32 DO							// Intel 80386 Ref Manual: "If a shift count greater than 31 is attempted, only the bottom five bits of the shift count are used."
                CODE(A.SHR, A0, T0)
            ELSE $(
				EMIT("movl @A,%ecx", A0, T0, 0, FALSE)
				EMIT("jcxz 1f")
				EMIT("2:")
				EMIT("decl %ecx")
				EMIT("shrl $1,%eax")
				EMIT("cmpl $32,%ecx")						// shift value still ge 32?
				EMIT("jge 2b")
				EMIT("shrl %cl,%eax")
				EMIT("1:")
            $)
            ENDCASE
        CASE 18:											// (18)	A := B LOGAND A
            CODE(A.AND, A0, T0)
            ENDCASE
        CASE 19:											// (19)	A := B LOGOR A
            CODE(A.OR, A0, T0)
            ENDCASE
        CASE 20:											// (20)	A := B NEQV A
            CODE(A.XOR, A0, T0)
            ENDCASE
        CASE 21:											// (21)	A := B EQV A
            EMIT("xorl $-1,%eax")
            CODE(A.XOR, A0, T0)
            ENDCASE
        CASE 22:											// (22)	FINISH
            EMIT("jmp finish")
            ENDCASE
        CASE 23:											// (23)	Switch on the value of A using data in the location
															//		addressed by C, C+1 etc.
															//		B, D := Location(C), Location(C + 1)
															//		UNTIL B = 0 DO
															//			$(	B, C := B - 1, C + 2
															//				IF A=Location(C) DO
															//					$(	D := Location(C + 1)
															//						BREAK $)       $)
															//		C := D
            EMIT("movl @A,%esi", XL, T.LL, 0, FALSE)
            EMIT("movl (%esi),%ecx")
            EMIT("movl 4(%esi),%edx")
            EMIT("jecxz 2f")
            EMIT("1:")
            EMIT("addl $8,%esi")
            EMIT("cmpl (%esi),%eax")
            EMIT("je 3f")
            EMIT("loop 1b")
            EMIT("2:")
            EMIT("jmp **%edx")
            EMIT("3:")
            EMIT("jmp **4(%esi)")
            L!LN := XL
            LN := LN + 1
            XL := XL + 1
            ENDCASE
        CASE 24:											// (24)-(35)	These are implementation dependent instructions
            EMIT("call selectinput")
            ENDCASE
        CASE 25:
            EMIT("call selectoutput")
            ENDCASE
        CASE 26:
            EMIT("call rdch")
            ENDCASE
        CASE 27:
            EMIT("call wrch")
            ENDCASE
        CASE 28:
            EMIT("call findinput")
            ENDCASE
        CASE 29:
            EMIT("call findoutput")
            ENDCASE
        CASE 30:
            EMIT("jmp stop")
            ENDCASE
        CASE 31:											// (31)	LEVEL()
            EMIT("movl (%ebp),%eax")
            ENDCASE
        CASE 32:											// (32)	LONGJUMP()
            EMIT("movl %ecx,%ebp")
            EMIT("jmp **%eax")
            ENDCASE
        CASE 33:											// (33)	ENDREAD()
            EMIT("call endread")
            ENDCASE
        CASE 34:											// (34)	ENDWRITE()
            EMIT("call endwrite")
            ENDCASE
        CASE 35:											// (35)	APTOVEC()
		//	Diagnostic code:
		//	SELECTOUTPUT(ERR)
		//	WRITEF("	F=%C,  A=%N,  T=%N,  IA=%N*n	F0=%C, A0=%N, T0=%N, IA0=%N*n", F, A, T, IA, F0, A0, T0, IA0)
		//	WRITEF("	F2=%C, A2=%N, T2=%N, IA2=%N*n	F3=%C, A3=%N, T3=%N, IA3=%N*n", F2, A2, T2, IA2, F3, A3, T3, IA3)
		//	SELECTOUTPUT(ASM)
		
		//	...preceding the APTOVEC() code we have:
		//	movl 12(%ebp),%eax
		//	movl %eax,%ebx						// Collect SIZE (in words)
		//	movl 8(%ebp),%eax					// Collect address of subroutine
		//	...then:
            EMIT("movl %ebp,%esi")				// collect current frame base
            EMIT("movl %eax,%ebx")
            EMIT("incl %ebx")					// increment SIZE
            EMIT("shll $2,%ebx")				// convert SIZE to bytes
            EMIT("addl %ebx,%esi")				// calculate new frame base address
            EMIT("movl (%ebp),%ebx")			// collect old frame base address in %ebx...
            EMIT("movl %ebx,(%esi)")			// ...and save in new frame base location
            EMIT("movl 4(%ebp),%ebx")			// collect old %eip (return address) value...
            EMIT("movl %ebx,4(%esi)")			// ...and save in new frame base
            EMIT("movl %ebp,%ebx")				// collect current frame base address...
            EMIT("shrl $2,%ebx")				// ...convert to word address...
            EMIT("movl %ebx,8(%esi)")			// ...and save in new stack frame
            EMIT("movl %eax,12(%esi)")			// Save requested SIZE (words) in new stack frame
            EMIT("movl %esi,%ebp")				// adjust frame base pointer for new function
            EMIT("jmp **%ecx")					// Call requested subroutine!
            ENDCASE
        CASE 36:											// (36)	GETBYTE()
            EMIT("shll $2,%eax")
            EMIT("addl %eax,%ecx")
            EMIT("movzbl (%ecx),%eax")
            ENDCASE
        CASE 37:											// (37)	PUTBYTE()
		//	Diagnostic code:
		//	SELECTOUTPUT(ERR)
		//	WRITEF("	F=%C,  A=%N,  T=%N,  IA=%N*n	F0=%C, A0=%N, T0=%N, IA0=%N*n", F, A, T, IA, F0, A0, T0, IA0)
		//	WRITEF("	F2=%C, A2=%N, T2=%N, IA2=%N*n	F3=%C, A3=%N, T3=%N, IA3=%N*n", F2, A2, T2, IA2, F3, A3, T3, IA3)
		//	WRITEF("	F4=%C, A4=%N, T4=%N, IA4=%N*n", F4, A4, T4, IA4)
		//	SELECTOUTPUT(ASM)
			EMIT("#		code?")
            TEST F2='S' THEN $(
        // args on the stack are as follows for ocode:
        //         top of stack  ->  arg1	= offset	@IA2
        //                           arg2	= vector	@IA2-1
        //                           arg3	= operand	@IA2-2
        // where the operation is  arg2%arg1 := arg3
                EMIT("movl @A,%ecx", IA2-1, T.LP, 0, FALSE)			// collect BCPL vector address
                EMIT("shll $2,%ecx")								// ...convert to real address
                EMIT("addl @A,%ecx", IA2, T.LP, 0, FALSE)			// add byte offset
                EMIT("movl @A,%eax", IA2-2, T.LP, 0, FALSE) $)		// collect operand
            ELSE TEST F2='L' THEN $(
        // args on the stack are as follows for iclib.i code:
        //         top of stack  ->  arg1	= operand	@IA3
        //                           arg2	= offset	(%ecx)
        //                           arg3	= vector	(%eax)
        // where the operation is  arg2%arg1 := arg3
                EMIT("shll $2,%eax")								// ...convert vector to real address
                EMIT("addl %eax,%ecx")								// add to byte offset
                EMIT("movl @A,%eax", IA3, T.LP, 0, FALSE) $)		// collect operand
			ELSE ERROR(21)

            EMIT("movb %al,(%ecx)")									// move operand to target
            ENDCASE
        CASE 38:
            EMIT("call input")
            ENDCASE
        CASE 39:
            EMIT("call output")
            ENDCASE
        CASE 40:
            EMIT("call unrdch")
            ENDCASE
        CASE 41:
            EMIT("call rewind")
			ENDCASE
        CASE 42:
            EMIT("call getvec")
            ENDCASE
        CASE 43:
            EMIT("call freevec")
            ENDCASE
		CASE 44:											// (44)	A := ABS A
            EMIT("testl %eax,%eax")
            EMIT("jg 1f")									// jump if positive...
            EMIT("negl %eax")								// ...otherwise negate it
            EMIT("1:")
            ENDCASE
        CASE 45:											// (45)	FINDAPPEND()
			EMIT("call findappend")
			ENDCASE
		CASE 46:											// (46)	Load selector - SELLD
		// SELLD takes two argments len and sh. Its effect is equivalent to
		// P!(S-1) := !(P!(S-1)) >> sh & mask
		// where mask is a bit pattern containing len right justified ones. If len is zero no
		// masking is done.

		// args on the stack are as follows:
        //         top of stack  ->  arg1	= shift		=IA2
        //                           arg2	= length	=IA4
        //                           arg3	= operand1	=IA6

		// ...so, that was easy.  Now just code it...

		CASE 47:											// (47)	Store selector - SELST
		// SELST takes three argments op, len and sh. If op is zero, its effect is equivalent to
		// SLCT len:sh:0 OF (P!(S-1)) := P!(S-2); S := S-2
		// but if op is non zero it represents an assignment operator (assop) and the statement
		// is equivalent to:
		// SLCT len:sh:0 OF (P!(S-1)) assop:= P!(S-2); S := S-2
		// The mapping between op and assop is given by the following table.
		
		//			op assop| op assop
		//			________|________
		//			 1   !  |  9   +
		//			 2  #*  | 10   -
		//			 3  #/  | 11  <<
		//			 4  #+  | 12  >>
		//			 5  #-  | 13   &
		//			 6   *  | 14   |
		//			 7   /  | 15  EQV
		//			 8  REM | 16  NEQV

		// The SELST operator with len and sh both zero is used in the compilation of
		// assop:= assignments where the left hand side is a simple variable or a subscripted
		// expression. For instance, the assigment v!3+:=1 might generate the following OCODE:
		// 		LG 200 LN 1 PLUS LN 1 SELST 9 0 0

		// args on the stack are as follows:
        //         top of stack  ->  arg1	= shift		=IA2
        //                           arg2	= length	=IA4
        //                           arg3	= op		=IA6
        //                           arg4	= operand1	=IA8
        //                           arg5	= operand2	=IA10

		//	Diagnostic code:
		//	SELECTOUTPUT(ERR)
		//	WRITEF("	F1=%C, A1=%N,	T1=%N*n", F1, A1, T1)
		//	WRITEF("	F=%C,  A=%N,	T=%N,  IA=%N*n	F0=%C, A0=%N,	T0=%N, IA0=%N*n", F, A, T, IA, F0, A0, T0, IA0)
		//	WRITEF("	F2=%C, A2=%N,	T2=%N, IA2=%N*n	F3=%C, A3=%N,	T3=%N, IA3=%N*n", F2, A2, T2, IA2, F3, A3, T3, IA3)
		//	WRITEF("	F4=%C, A4=%N,	T4=%N, IA4=%N*n	F5=%C, A5=%N,	T5=%N, IA5=%N*n", F4, A4, T4, IA4, F5, A5, T5, IA5)
		//	WRITEF("	F6=%C, A6=%N,	T6=%N, IA6=%N*n	F7=%C, A7=%N,	T7=%N, IA7=%N*n", F6, A6, T6, IA6, F7, A7, T7, IA7)
		//	WRITEF("	F8=%C, A8=%N,	T8=%N, IA8=%N*n	F9=%C, A9=%N,	T9=%N, IA9=%N*n", F8, A8, T8, IA8, F9, A9, T9, IA9)
		//	WRITEF("	F10=%C, A10=%N,	T10=%N, IA10=%N*n*n", F10, A10, T10, IA10)
		//	SELECTOUTPUT(ASM)
			$(	LET MASK=-1									// mask of all ones
				IF IA4=0 IA4:=BITSPERWORD					// (a 'length' of zero means full word mask)
				UNLESS IA4=BITSPERWORD						// if 'length' is full word cut to the chase, otherwise...
					MASK := MASK >> (BITSPERWORD-IA4)		// required 'length' mask now at rh end
				UNLESS IA2=0
					MASK := MASK << IA2						// Apply the 'shift' to the mask
						EMIT("#		@S code...", A=47->"SELST","SELLD")
						EMIT("movl @A,%eax", A=47->IA7,IA5, T.LP, 0, FALSE)		// collect (BCPL pointer to) operand1 from stack
						EMIT("movl (,%eax,4),%eax")								// collect value of operand1
				TEST A=47 THEN $(												// for SELST...
						EMIT("movl @A,%ecx", IA10, T10=T.R->T.N,T.LP, 0, FALSE)	// collect operand2 from stack
					SWITCHON IA6 INTO $(					// now apply the operator
					CASE  0:								// :=
					UNLESS IA2=0
						EMIT("shll @A,%ecx", IA2, T.N, 0, FALSE)				// apply the 'shift'
						EMIT("xorl %ecx,%eax")
					UNLESS MASK=-1
						EMIT("andl @A,%eax", MASK, T.N, 0, FALSE)				// apply the prescribed MASK, unless it's all ones
						EMIT("movl @A,%ecx", IA7, T.LP, 0, FALSE)				// collect (BCPL pointer to) operand1 from stack
						EMIT("xorl (,%ecx,4),%eax")								// collect value of operand1
						ENDCASE
					CASE  6:								// *
						EMIT("imull %ecx,%eax")							// multiply operand1 by operand2
						ENDCASE
					CASE  7:								// /
					CASE  8:								// REM
						EMIT("cltd")
						EMIT("idivl %ecx")								// divide operand1 by operand2
						IF IA6=8
							EMIT("movl %edx,%eax")						// pass REMainder as result
						ENDCASE
					CASE  9:								// +
						EMIT("addl %ecx,%eax")							// add operand2 to masked & shifted operand1
						ENDCASE
					CASE 10:								// -
						EMIT("subl %ecx,%eax")							// subtract operand2 from masked & shifted operand1
						ENDCASE
					CASE 11:								// <<
						EMIT("shll %cl,%eax")							// left shift operand1 by operand2	(NB: see LSHIFT above)
						ENDCASE
					CASE 12:								// >>
						EMIT("shrl %cl,%eax")							// right shift operand1 by operand2	(NB: see RSHIFT above)
						ENDCASE
					CASE 13:								// &
						EMIT("andl %ecx,%eax")							// logical and operand1 with operand2
						ENDCASE
					CASE 14:								// |
						EMIT("orl %ecx,%eax")							// logical or operand1 with operand2
						ENDCASE
					CASE 15:								// EQV
						EMIT("xorl $-1,%eax")							// bitwise equivalence of operand1 with operand2
						EMIT("xorl %ecx,%eax")
						ENDCASE
					CASE 16:								// NEQV
						EMIT("xorl %ecx,%eax")							// bitwise non-equivalence of operand1 with operand2
						ENDCASE
					CASE  1:								// !
																		// vecap, apparently.
					CASE  2:								// #*
					CASE  3:								// #/
					CASE  4:								// #+
					CASE  5:								// #-
					DEFAULT:
						ERROR(22)										// Unsupported operator
						ENDCASE
					$)
						EMIT("movl @A,%ecx", IA7, T.LP, 0, FALSE)		// collect (BCPL pointer to) operand1 from stack
						EMIT("shll $2,%ecx")							// real pointer to operand1
						EMIT("movl %eax,(%ecx)")						// save result...
				$) ELSE $(												// for SELLD...
				UNLESS MASK=-1
						EMIT("andl @A,%eax", MASK, T.N, 0, FALSE)		// apply the prescribed MASK, unless it's all ones
				UNLESS IA2=0
						EMIT("shrl @A,%eax", IA2, T.N, 0, FALSE)		// 'shift' result to rh end of %eax
						CODE(A.MV2, IA5, T.LP)							// save result...
				$)
						EMIT("#		...")
			$)
			ENDCASE
        CASE 49:											// (49)	RENAME()
			EMIT("call rename")
			ENDCASE
        CASE 50:											// (50)	DELETE()
			EMIT("call delete")
			ENDCASE
        CASE 51:											// (51)	GETENV()
			EMIT("call getenv")
			ENDCASE
        CASE 54:											// (54)	FINDFILE()
			EMIT("call findfile")
			ENDCASE
        $)
    $)
	F10, A10, T10, IA10 := F9, A9, T9, IA9
	F9, A9, T9, IA9 := F8, A8, T8, IA8
	F8, A8, T8, IA8 := F7, A7, T7, IA7
	F7, A7, T7, IA7 := F6, A6, T6, IA6
	F6, A6, T6, IA6 := F5, A5, T5, IA5
	F5, A5, T5, IA5 := F4, A4, T4, IA4
	F4, A4, T4, IA4 := F3, A3, T3, IA3
	F3, A3, T3, IA3 := F2, A2, T2, IA2
    F2, A2, T2, IA2 := F0, A0, T0, IA0
    F0, A0, T0, IA0 := F,  A,  T,  IA
    F,  A,  T,  IA  := F1, A1, T1, 0
$)

AND EPILOG() BE
$(  SECT := 1
    EMIT(".global G")
    EMIT(".align 4")
    EMIT("G:")
    FOR I = 0 TO GSZ-1
        EMIT(".long @A # @N", G!I, G!I -> T.LL, T.N, I, 1)
$)

AND CODE(OP, A, T) BE CODE1(OP, A, T, 0)
AND CODE1(OP, A, T, R) BE EMIT(ASTR(OP), A, T, R, OP<=A.JMP)
AND DATA(S, A, T) BE EMIT(S, A, T, 0, TRUE)

AND ASTR(X) = VALOF
    SWITCHON X INTO $(
    CASE A.JE:  RESULTIS "je @A"
    CASE A.JNE: RESULTIS "jne @A"
    CASE A.JL:  RESULTIS "jl @A"
    CASE A.JGE: RESULTIS "jge @A"
    CASE A.JG:  RESULTIS "jg @A"
    CASE A.JLE: RESULTIS "jle @A"
    CASE A.JMP: RESULTIS "jmp @A"
    CASE A.MUL: RESULTIS "imull @A"
    CASE A.DIV: RESULTIS "idivl @A"
    CASE A.MOV: RESULTIS "movl @A,@R"
    CASE A.ADD: RESULTIS "addl @A,@R"
    CASE A.SUB: RESULTIS "subl @A,@R"
    CASE A.CMP: RESULTIS "cmpl @A,@R"
    CASE A.SHL: RESULTIS "shll @A,@R"
    CASE A.SHR: RESULTIS "shrl @A,@R"
    CASE A.AND: RESULTIS "andl @A,@R"
    CASE A.OR:  RESULTIS "orl @A,@R"
    CASE A.XOR: RESULTIS "xorl @A,@R"
    CASE A.MV2: RESULTIS "movl @R,@A"
    CASE A.LEA: RESULTIS "leal @A,@R"
	CASE A.SL2: RESULTIS "shll %cl,@R"
	CASE A.SR2: RESULTIS "shrl %cl,@R"
    CASE A.SNE: RESULTIS "setne %al"
    CASE A.SE:  RESULTIS "sete %al"
    CASE A.SGE: RESULTIS "setge %al"
    CASE A.SL:  RESULTIS "setl %al"
    CASE A.SLE: RESULTIS "setle %al"
    CASE A.SG:  RESULTIS "setg %al"
    DEFAULT: ERROR(9)
    $)

AND EMIT(S, A, T, X, J) BE
$(  STATIC $( PSECT=0 $)
    UNLESS SECT=PSECT $(
        WRITES(SECT=0 -> ".text", ".data"); WRCH('*N')
        PSECT := SECT $)
    IF LN $(
        WRITES(".align 4*N")
        FOR I = 0 TO LN - 1 $( WRCH('L'); WRN(L!I); WRITES(":*N") $)
        LN := 0 $)
    FOR I = 1 TO GETBYTE(S, 0) $(
        LET C = GETBYTE(S, I)
        TEST C='@' $(
            I := I + 1
            SWITCHON GETBYTE(S, I) INTO $(
            CASE 'A':
                TEST NOT J & (T=T.N | T=T.LL) WRCH('$')
                OR IF J & T>=T.R WRCH('**')
                ARGOUT(A, T)
                ENDCASE
            CASE 'N':
                ARGOUT(X, T.N)
                ENDCASE
            CASE 'S':							// Output a string					DC: 7-August-2013
                WRITEF("%S", A)
                ENDCASE
            CASE 'R':
                ARGOUT(X, T.R)
            $)
         $) OR WRCH(C)
    $)
    WRCH('*N')
$)

AND ARGOUT(A, T) BE
$(  TEST T=T.R | T=T.IR $(
        IF T=T.IR WRITES("(,")
        WRITES(A=0 -> "%eax", "%ecx")
        IF T=T.IR WRITES(",4)")
    $) OR $(
        LET K, E = T & 3, K=T.LP | K=T.LG
        IF K=T.LL | K=T.LF | T=T.LF WRCH('L')
        IF E DO A := A * WORDSZ
        WRN(A)
        IF E WRITES(K=T.LP -> "(%ebp)", "(%edi)")
    $)
$)

AND WRN(N) BE
$(  TEST N<0 WRCH('-') OR N := -N
    IF N<-9 WRN(-N / 10)
    WRCH('0' - N REM 10)
$)

AND RDN() = VALOF
$(  LET X, N, I = 0, FALSE, 0
    IF CH='-' $( N := TRUE; RCH() $)
    WHILE '0' <= CH <= '9' $(
        X := X * 10 - (CH - '0')
        I := I + 1
        RCH()
    $)
    IF I=0 ERROR(10)
    RESULTIS N -> X, -X
$)

AND RCH() BE
$(  LET N=0
    RCH1()
    UNLESS CH='/' RETURN
    $( RCH1();
       N := N+1
       IF CH=ENDSTREAMCH DO ERROR(11)
       PUTBYTE(RN, N, CH)			// Assemble any descriptive text
    $) REPEATUNTIL CH='*N'
	PUTBYTE(RN, 0, N-1)
$) REPEAT

AND RCH1() BE
$(  IF CH='*N' LINE, COL := LINE + 1, 0
    COL := COL + 1
    CH := RDCH()
$)

AND ERROR(N) BE
$(	LET S = "?"
	SELECTOUTPUT(ERR)
    UNLESS LINE=0
        WRITEF("stdin(%N,%N): ", LINE, COL)
		
		SWITCHON N INTO $(
		DEFAULT:	S := "Unknown Error code";						ENDCASE
		CASE 1:		S := "ASSEM() called with illegal argument";	ENDCASE
		CASE 2:		S := "Global number beyond end of GVEC array";	ENDCASE
		CASE 3:		S := "G found with modifier other than 'L'";	ENDCASE
		CASE 4:		S := "Literal value found where none expected";	ENDCASE
		CASE 5:		S := "Label number beyond end of LVEC array";	ENDCASE
		CASE 6:		S := "Unconditional jump to non-static";		ENDCASE
		CASE 7:		S := "K found with non-literal value";			ENDCASE
		CASE 8:		S := "X found with out-of-range value";			ENDCASE
		CASE 9:		S := "ASTR() called with illegal argument";		ENDCASE
		CASE 10:	S := "Attempt to read number finds NIL";		ENDCASE
		CASE 11:	S := "Premature end of INTCODE input file";		ENDCASE

		CASE 20:	S := "SECTION label exceeds character limit";	ENDCASE
		CASE 21:	S := "Problem identifying value for a%i:=...";	ENDCASE
		CASE 22:	S := "Unsupported operator for SELST instruction";	ENDCASE
		$)

	WRITEF("xg error %N: %S*N", N, S)
    STOP(1)
$)
