@echo off

rem	Copyright (c) 2013-2016 Dave Cannon.  All rights reserved.
rem	$Id: bcpl.cmd,v 2.06 11-Jan-2016 09:24:20 rn Exp $

setlocal enabledelayedexpansion
if errorlevel 1 (
  echo Unable to enable delayed expansion - aborting!
  goto eof
)

set vn=2
set rvn=06
set edt=11-Jan-2016

set version=%0 Vn %vn%.%rvn%	%edt%

rem	Vn 2.06	11-Jan-2016 change order of concatenation of INTCODE files so 'xg_i'
rem						generates appropriate/helpful diagnostics
rem	Vn 2.05	20-Aug-2014 all environment variables now set externally by ..\BCPLenv.cmd
rem	Vn 2.04	24-Aug-2013 introduction of 'getopt' to parse command line options
rem	Vn 2.03	09-Feb-2013 BCPLHDRS environment variable added as location of 'libhdr', etc.
rem	Vn 2.02	05-Feb-2013 adapted to run the 1-Feb-2011 version of the BCPL compiler
rem	Vn 2.01	07-Jan-2013 Employs xg Vn 2.03 developed to translate INTCODE to assembler
rem	Vn 2.00	06-Jan-2013	modified to run Martin Richards' latest compiler version

rem	Vn 1.05	26-Oct-2012	add '-c' option to preserve the OCODE output as'file.ocd'
rem	Vn 1.04	12-Oct-2012	add '-i' option to preserve the INTCODE output as'file.i'
rem	Vn 1.03	09-Oct-2012	add '-l' option to preserve the 'file.o' output for ld
rem	Vn 1.02	29-Sep-2012	add '-S' option to preserve the 'file.asm' output
rem	Vn 1.01	28-Sep-2012	add quotes to variable 'd' for compatibility with Windows XP
rem	Vn 1.00	27-Sep-2012	Initial translation from Robert Nordier's shell script to DOS

if not '%1'=='' goto doit

:usage
echo	Usage:	%0 [-Options] [-o output] file
echo		where	file is a BCPL source program to be compiled
echo			[-o output] (optional) specifies the compiled program name
echo			[-Options] may be any of:
echo			[-O] (optional) invokes the copt code optimiser
echo			[-c] (optional) preserve OCODE output as 'file.ocd'
echo			[-i] (optional) preserve INTCODE output as 'file.i'
echo			[-l] (optional) preserve the 'file.o' output for ld
echo			[-S] (optional) preserve the 'file.asm' output
echo.
echo		%version%
echo.
goto eof

:doit

set oflag=0
set args=

if not defined GnuWin (
  echo.
  echo 	The BCPL environment variables have not been established.
  echo 	You must execute ..\BCPLenv.cmd before running this script
  echo.
  goto eof
)

set LDFLAGS=-Bdynamic -T"%GnuWin%\mingw32\lib\ldscripts\i386pe_mgw.x" "%GnuWin%\lib\crt2.o" "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtbegin.o"
set LDLIBS=-lmingw32 -lmoldname -lmingwex -lmsvcrt -luser32 -lkernel32 -ladvapi32 -lshell32 "%GnuWin%\lib\gcc\mingw32\%minGWVn%\crtend.o"
set BCPLvn=%BCPLroot%\BCPL 2014-08-13

for /f "usebackq tokens=*" %%i in (`getopt cilOSo: %*`) do set args=%%i
for %%i in (%args%) do (
	if '!oarg!'=='?' (
	  set oarg=%%i
	)
	if '!farg!'=='?' (
	  set farg=%%i
	)
	if '%%i'=='-O' (
	  set oflag=1
	)
	if '%%i'=='-o' (
	  set oarg=?
	)
	if '%%i'=='-c' (
	  set ocd=1
	)
	if '%%i'=='-i' (
	  set int=1
	)
	if '%%i'=='-l' (
	  set obj=1
	)
	if '%%i'=='-S' (
	  set asm=1
	)
	if '%%i'=='--' (
	  set farg=?
	)
)
if '%farg%'=='?' goto usage
if '%oarg%'=='?' goto usage

if not exist %farg% (
	echo "%0 %farg%: no such file"
	goto usage
)

if exist OCODE del /q/f OCODE > nul
if exist INTCODE del /q/f INTCODE > nul
if exist ASM del /q/f ASM > nul

rem if not exist libhdr cp %BCPLsrc%\libhdr+ libhdr

if defined oarg (
    set f=%oarg%
) else (
	for /f "usebackq" %%i in (`basename %farg% .b`) do set f=%%i
)

if exist %f%.ocd del /q/f %f%.ocd > nul
if exist %f%.int del /q/f %f%.int > nul
if exist %f%.asm del /q/f %f%.asm > nul
if exist %f%.s del /q/f %f%.s > nul
if exist %f%.o del /q/f %f%.o > nul

IF defined ocd (
	"%BCPLvn%\bcpl+" %farg%
	move OCODE %f%.ocd > nul
) else (
	"%BCPLvn%\bcpl+" %farg% %f%.i
	cat %f%.i "%BCPLlib%\blib.i" "%BCPLlib%\iclib.i" > %f%.int
	if defined int (
		move %f%.i %f%.int > nul
		goto eof
	) else (
		del /q/f %f%.i > nul
	)
	"%BCPLvn%\codegens\xg_i" < %f%.int
	if %oflag% equ 1 (
		"%BCPLcopt%\copt" "%BCPLsrc%\rules" < ASM > ASM.op && mv ASM.op ASM
	)
	move ASM %f%.s > nul
	as -o %f%.o %f%.s
	if not defined obj if exist %f%.o ld %LDFLAGS% -o %f%.exe %f%.o "%BCPLsrc%\su.o" "%BCPLsrc%\rtc.o" %LDLIBS%
)

if exist %f%.int del /q/f %f%.int > nul
if defined asm (
	if exist %f%.s move %f%.s %f%.asm > nul
	)
if exist %f%.s del /q/f %f%.s > nul
if not defined obj if exist %f%.o del /q/f %f%.o > nul

:eof

set f=
set vn=
set rvn=
set edt=
set int=
set ocd=
set version=
set oflag=
set asm=
set obj=
set oarg=
set args=
set LDFLAGS=
set LDLIBS=
