// $Id: xg.b,v 4.08 2021/05/08 15:19:21 rn Exp $

// Copyright (c) 2004, 2020 Robert Nordier.  All rights reserved.
//
// BCPL compiler backend: generates x86 assembler from INTCODE


// Amended to accept INTCODE from Martin Richards' current BCPL compiler
//				 								2013-2021 Dave Cannon


GET "LIBHDR"

MANIFEST $(
	codename	=TRUE							// Enable/disable procedure naming for run-time diagnostics
	G_codebase	=-2
	G_codetop	=-1
$)

MANIFEST $( WORDSZ=4; GSZ=maxglobal; LSZ=10000	// GSZ originally 500, LSZ originally 500
			LINC=10000 $)						// LINC originally 500				DC: 29-March-2013

// Type values (N=number, P=local, G=global, L=static, R=register)
// at up to two levels of indirection
MANIFEST $(
T_I  =#X04										// The indirect bit, set at *** below (which may become...
T_II =#X08										// ...the 'Store' bit

T_N  =#X00
T_LP =#X01
T_LG =#X02
T_LL =#X03

T_R  =#X04										// = T_N  | T_I
T_P  =#X05										// = T_LP | T_I
T_G  =#X06										// = T_LG | T_I
T_L  =#X07										// = T_LL | T_I

T_IR =#X08										// = T_N  | T_II
T_IP =#X09										// = T_LP | T_II
T_IG =#X0A										// = T_LG | T_II
T_IL =#X0B										// = T_LL | T_II

T_LF =#X10										// T_LF added						DC:	 6-August-2013
T_LE =#X11										// T_LE 	load @EXTERNAL			DC:	10-May-2020

T_SE =#X15										// T_SE		store EXTERNAL
T_IE =#X15										// T_IE		load  EXTERNAL			DC:  2-July-2020

T_SIE=#X19
$)

MANIFEST $(
  UPB=11										// Maximum SECTION string length	DC: 7-August-2013
  extlim=20										// Maximum characters in EXTERNAL name
  codesizemax=5000								// Bytes:  this needs to be the same as 'codesizemax' in bcplfecg.h
$)

// Assembler statement patterns for code generation: with the
// jump instructions first, as they need to be distinguished
MANIFEST $(
A_JE=0;   A_JNE=1;  A_JL=2;   A_JGE=3;  A_JG=4;   A_JLE=5;  A_JMP=6
A_MUL=7;  A_DIV=8;  A_MOV=9;  A_ADD=10; A_SUB=11; A_CMP=12; A_SHL=13
A_SHR=14; A_AND=15; A_OR=16;  A_XOR=17; A_MV2=18; A_LEA=19; A_SL2=20
A_SR2=21; A_SNE=22; A_SE=23;  A_SGE=24; A_SL=25;  A_SLE=26; A_SG=27

// ...and now the floating-point instructions										DC:20-November-2013
A_FML=30; A_FDV=31;           A_FAD=33; A_FSB=34;
A_FNE=35; A_FE=36;  A_FGE=37; A_FL=38;  A_FLE=39; A_FG=40
$)

GLOBAL $(
G		: 150			// Definitions vector for globals
L		: 151			// Definitions vector for labels
LN		: 152			// Labels in L vector
LOFF	: 153			// Label offset value
SECT	: 154			// Section (text or data) in output
LINE	: 155			// Source line number
COL		: 156			// Source column number
ch		: 157			// Source character last read
ERR		: 158			// Stream for errors
ASM		: 159			// Stream for output
S		: 160			// Container for SECTION name
RN		: 161			// Container for descriptive text
extlist	: 162			// Definitions chain of EXTERNALs
$)

MANIFEST $(				// offsets in extlist member defining this EXTERNAL
e_link	= 0
e_type	= 1				// may be one of:	s_datalab, s_iteml, s_entry
e_lab	= 2
e_flag	= 3
e_name	= 4

e_wsize	= e_name

s_entry		= 120		// ...from "..\include\bcplfecg.h":		change that and these MUST also be changed!
s_datalab	= 125
s_iteml		= 132
$)

STATIC $(
instr		= 0			// count instructions assembled
codecount	= TRUE
$)

LET start() = VALOF
$(  LET GVEC = getvec(GSZ+neggsz)
    AND LVEC = getvec(LSZ)
	AND STR = getvec(codesizemax/bytesperword)
	AND RNAM = getvec(256/bytesperword)

    G := GVEC
    L := LVEC
	S := STR
	RN := RNAM
	putbyte(RN, 0, 0)
	extlist := 0

    FOR I = 0 TO GSZ+neggsz DO G!I := 0
	G := GVEC+neggsz			// G now allows neggsz negative values...

    LN, SECT, LINE, COL := 0, 0, 0, 0
    LOFF := 20000				// Offset (originally 2000) to add to intcode label number - incremented by LINC on new segment of code
    ch := '*N'
    ERR := OUTPUT()
    ASM := FINDOUTPUT("ASM")
    selectoutput(ASM)

	emit("# Code generated by XG Vn 4.08 (8-May-2021)*n")
    IF codename $(
      emit("	.global	code_base")
	  emit("	code_base:")
	$)
    assem()
    epilog()
    ENDWRITE()
//	writef("XG assembler Vn 4.08: Code = %n lines*n", instr)
    RESULTIS 0
$)

AND rdtext(F) = VALOF
$(	LET lim, diag, A = UPB, 20, ?
	LET txt=getvec(codesizemax/bytesperword)
	
	SWITCHON F INTO $(
	  CASE 'E':								// EXTERNAL name
		lim := extlim
		diag := 24	  
		ENDCASE
	  CASE 'M':								// Assembler ('M'achine) code text
		lim := codesizemax					// Bytes:  this needs to be the same as 'codesizemax' in bcplfecg.h
		diag := 23
		ENDCASE
	  CASE 'N':								// SECTION label
		lim := UPB
		diag := 20	  
		ENDCASE
	$)
		
	rch()
	A := rdn()								// String length
	putbyte(txt, 0, A <= lim -> A, lim)
	FOR I=1 TO A DO $(
	  IF I > lim error(diag, txt, lim)
	  rch()
	  SWITCHON ch INTO $(
		CASE 'C':							// encoded text
		  rch()
		  IF ch='C' rch()					//	discard leading 'C'
		  putbyte(txt, I, rdn())
//	fwritef(syserr, "I=%n, A=%n, ch='%c'*n", I, A, txt%I)
		ENDCASE
		  
		CASE '*"':							// quoted text
		  $(
			rch()
			putbyte(txt, I, ch)
			I := I+1
		  $) REPEATUNTIL I>A
		  rch(); rch()						//	discard trailing '" '
		ENDCASE

		DEFAULT:
		  error(diag)
	  $)
	$)
	RESULTIS txt
$)

AND fwritef(strm, fmt, a, b, c, d, e, f, g, h, i, j, k) BE
$(	LET cos=output()
	selectoutput(strm)
	writef(fmt, a, b, c, d, e, f, g, h, i, j, k)
	selectoutput(cos)
$)

/*
 *AND binaryw(w, v) = VALOF
 *$( LET bit31=#x80000000
 *   AND i, j = 32, 1
 *   $( LET bit=w&bit31
 *	 fwritef(syserr, "i=%n, j=%n, bit31=x%x8*n", i, j, bit31)
 *      v%j := bit ~= 0 -> '1', '0'
 *	  bit31 := bit31>>1						// bit31 becomes bit30, etc...
 *	  j := j+1
 *	  i := i-1
 *   $) REPEATUNTIL i=0
 *   v%0 := 32
 *   RESULTIS v
 *$)
 */

AND assem() BE
$(  LET F, A, T, I = 0, 0, 0, 0
    rch()
    F := ch
    SWITCHON F INTO $(
    DEFAULT:
        error(1)
    CASE endstreamch:
        gener(0, 0, 0)
        RETURN
    CASE '*S': CASE '*N':
        LOOP
	CASE '$':									// Signals start of new routine, which may be labelled...	DC:	25-Oct-2013
		ENDCASE
	CASE 'E': $(								// EXTERNAL declaration										DC: 12-May-2020
		LET ev = getvec(e_wsize)				// container for this EXTERNAL
		rch();	ev!e_type	:= rdn()
		rch();	ev!e_lab	:= LOFF + rdn()
		ev!e_flag	:= 0						// flag
		T := rdtext(F)
		ev!e_name	:= T						// (pointer to) name of EXTERNAL
//	fwritef(syserr, "%s: %n L%n (%n) %s*n", "external", ev!e_type, ev!e_lab, ev!e_flag, ev!e_name)
		ev!e_link := extlist					// add to...
		extlist := ev							// ... chain of EXTERNAL declarations
		ENDCASE
		$)
	CASE 'M':									// Inline assembler ('M'achine) code						DC: 19-Dec-2015
	CASE 'N':									// SECTION label											DC: 7-August-2013
		T := rdtext(F)
		ENDCASE
    CASE 'Z':									// end of code section;  clear accumulated data
		UNTIL extlist=0 DO $(
		  LET nxt = extlist!e_link
		  freevec(extlist!e_name)
		  freevec(extlist)
		  extlist := nxt
		$)
		putbyte(RN, 0, 0)						// Clear any descriptive text
		codecount := FALSE
        LOFF := LOFF + LINC
        LOOP
    CASE 'G':
        rch()
        A := rdn()
        UNLESS A<GSZ error(2)
        UNLESS ch='L' error(3)
        rch()
        G!A := LOFF + rdn()
        LOOP
    CASE '0': CASE '1': CASE '2': CASE '3': CASE '4':
    CASE '5': CASE '6': CASE '7': CASE '8': CASE '9':
        F := 'B'
        A := LOFF + rdn()
        ENDCASE
    CASE 'A': CASE 'F': CASE 'J': CASE 'K':
    CASE 'L': CASE 'S': CASE 'T': CASE 'X':
	CASE 'W':									// Floating-point operation			DC: 21-Nov-2013
        rch()
        IF ch='I' $( I := 1; rch() $)
        T := ch='G' -> T_LG,
             ch='P' -> T_LP,
             ch='L' -> T_LL,
             ch='E' -> T_LE,					// load EXTERNAL					DC: 10-May-2020
			 ch='F' -> T_LF, T_N
        IF F='S' $( IF T=T_N error(4); I := I + 1 $)
		TEST T=T_LE THEN $(
		  A := rdtext('E')
		$) ELSE $(
          IF T rch()
		  A := rdn()
		  IF T=T_LL | T=T_LF A := A + LOFF
		$)
		T := I << 2 | T							// *** This makes some assumptions about the value of T and I!
//	IF F='S' fwritef(syserr, "...in assem(): CASE 'A'...'W' - T=%n (%s)*n", T, T=T_SE->"T_SE","T_other")
        ENDCASE
    CASE 'C':
        rch()
        A := rdn()
        ENDCASE
    CASE 'D':
        rch()
        IF ch='L' $( T := T_LL; rch() $)
        A := rdn()
        IF T=T_LL A := A + LOFF
    $)
    gener(F, A, T)
$) REPEAT

AND gener(F1, A1, T1) BE
$(  STATIC $(
		F=0;  A=0;  IA=0;  T=0			// previous parameter values...
		F0=0; A0=0; IA0=0; T0=0			// ...etc
		F2=0; A2=0; IA2=0; T2=0
		F3=0; A3=0; IA3=0; T3=0
		F4=0; A4=0; IA4=0; T4=0
		F5=0; A5=0; IA5=0; T5=0
		F6=0; A6=0; IA6=0; T6=0
		F7=0; A7=0; IA7=0; T7=0
		F8=0; A8=0; IA8=0; T8=0
		F9=0; A9=0; IA9=0; T9=0
		F10=0; A10=0; IA10=0; T10=0
		XL=0
	$)

    LET xdyadic(X) = 5<=X<=21 | X=32 | 35<=X<=37 |
						X=49 | X=54				// ...or RENAME() or FINDFILE()		DC: 22-Nov-2013

	AND debug(F1, A1, T1) BE
	$(	LET cos = output()
		selectoutput(syserr)
		writef("	F1=%C, A1=%N,	T1=%N*n", F1, A1, T1)
		writef("	F=%C,  A=%N,	T=%N,  IA=%N*n	F0=%C, A0=%N,	T0=%N, IA0=%N*n", F, A, T, IA, F0, A0, T0, IA0)
		writef("	F2=%C, A2=%N,	T2=%N, IA2=%N*n	F3=%C, A3=%N,	T3=%N, IA3=%N*n", F2, A2, T2, IA2, F3, A3, T3, IA3)
		writef("	F4=%C, A4=%N,	T4=%N, IA4=%N*n	F5=%C, A5=%N,	T5=%N, IA5=%N*n", F4, A4, T4, IA4, F5, A5, T5, IA5)
		writef("	F6=%C, A6=%N,	T6=%N, IA6=%N*n	F7=%C, A7=%N,	T7=%N, IA7=%N*n", F6, A6, T6, IA6, F7, A7, T7, IA7)
		writef("	F8=%C, A8=%N,	T8=%N, IA8=%N*n	F9=%C, A9=%N,	T9=%N, IA9=%N*n", F8, A8, T8, IA8, F9, A9, T9, IA9)
		writef("	F10=%C, A10=%N,	T10=%N, IA10=%N*n*n", F10, A10, T10, IA10)
		selectoutput(cos)
	$)

    AND xforce(X, T) =
        (5<=X<=7 & T=T_N) |
        (16<=X<=17 & T NE T_N) |
        X>=32

    AND load(R, J, FORCE) BE
    $(  LET E = T=T_LP | T=T_LG | T=T_LF | T=T_LE
        AND M = E | (NOT J & (T=T_LL | T=T_LF))
    //	AND I = T>=T_IR
		AND I = (T_IR<=T<=T_IL) | T=T_SIE
        IF FORCE | M | I $(
//	IF T&T_LE = T_LE fwritef(syserr, "...in load() calling code1(OP=%s, A=%n, T=%s, R=%n)*n", E->"A_LEA","A_MOV", A, T=T_LE->"T_LE","T_IE", R)
            code1(E -> A_LEA, A_MOV, A, T, R)
            IF M & NOT J & NOT T=T_LF code1(A_SHR, 2, T_N, R)
            A, T := R, I -> T_IR, T_R
        $)
    $)
	
	AND scanexterns(s) = VALOF
	$(	LET scmp(s1, s2) = VALOF
		$( 	LET l = s1%0
			FOR i=0 TO l DO $(
				UNLESS s1%i = s2%i RESULTIS 0
			$)
			RESULTIS s2
		$)
		LET ent, r = extlist, ?
		UNTIL ent=0 DO $(
			r := scmp(s, ent!e_name)
			IF r RESULTIS ent			// return the vec structure associated with this EXTERNAL
			ent := ent!e_link
		$)
		RESULTIS 0						// all EXTERNALs compared, no match found
	$)

    LET X = ?
    IA := A								// Preserve initial value of A		DC: 21-Oct-2013

    SWITCHON F INTO $(
    CASE 0:
        ENDCASE
	CASE '$':							// Entry point to new routine...
		UNLESS getbyte(RN, 0)=0 $(
            SECT := 1
		//	TEST scanexterns(RN) THEN $(				// this routine has been declared EXTERNAL, and is defined locally
		//		emit("	*n.align 4")
		//		SECT := 0
		//		emit("	__@S:", RN)
		//	$) ELSE $(					// it's a local routine (which may be declared as a GLOBAL) - we don't need the name as a label...
				SECT := 0
				emit("#	@S", RN)		// name the routine
				IF codename DO $(						// add the name of the routine to the code...
				  emit(".align 4")
				  emit("	.ascii	*"ptxt*"")				// ...with a prior tag
				  emit("	.asciz	*"@S*"", RN)			// NB:	this is NOT a BCPL string!
				$)
		//	$)
            putbyte(RN, 0, 0)			// ...clear the name
		$)
		ENDCASE
    CASE 'A':
	CASE 'S':
        IF T=T_SE | T=T_SIE $(					// store EXTERNAL
//	IF F='S' fwritef(syserr, "...in gener(): CASE 'S' - A=%n(%s), T=%s*n", A, T=T_SE|T=T_SIE->A,"", T=T_SE->"T_SE",T=T_SIE->"T_SIE","Other")
		//	emit("	movl %eax,@S", A)
		//	ENDCASE
		$)
        load(1, FALSE, FALSE)
        code((F='A' -> A_ADD, A_MV2), A, T)
        ENDCASE
	CASE 'B':
        UNLESS LN<LSZ error(5)
        L!LN := A
        LN := LN + 1
        ENDCASE
    CASE 'C':
        SECT := 1
        data("	.byte @A", A, 0)
        UNLESS F1='C' $(
			emit(".align 4,0")
		$)
        SECT := 0
        ENDCASE
    CASE 'D': 
        SECT := 1
		data("	.long @A", A, T)
        SECT := 0
        ENDCASE
    CASE 'E': $(						// DC: 10-May-2020		- EXTERNAL declaration
		LET ent = scanexterns(T)
		IF ent & (ent!e_type=s_entry) $(				// output data describing EXTERNAL entry to a procedure
			SECT := 1
			emit("	*n.align 4")
			emit("	.globl @S", T)
			emit("	@S:", T)				// routine entry point name		- DC: 10-May-2020
			emit("		.long	L@N", 0, 0, ent!e_lab)
			SECT := 0   $)
        ENDCASE
		$)
    CASE 'J':
        load(1, TRUE, FALSE)
        code(A_JMP, A, T)
        ENDCASE
    CASE 'K':
        UNLESS T=T_N $(
			selectoutput(ERR)
			writef("	F=%C,  A=%N,  T=%N*n	F1=%C, A1=%N, T1=%N*n", F, A, T, F1, A1, T1)
			error(7)
			$)
        emit("	movl %ebp,%ecx")
        emit("	addl @A,%ebp", A << 2, T_N, 0, FALSE);
        emit("	movl %ecx,(%ebp)")
        emit("	movl $1f,4(%ebp)")
        emit("	jmp **%eax")
        emit("	1:")
        ENDCASE
    CASE 'L':
//	fwritef(syserr, "...in gener(): CASE L - A=%n(%s), T=%s*n", A, T=T_LE->A,"", T=T_LE->"T_LE",T=T_IE->"T_IE","DEFAULT")
        SWITCHON T INTO $(
          DEFAULT:
            X := (F1='X' | F1='W') & xdyadic(A1) -> 1, 0				// DC: 22-Nov-2013		- 'W' for floating-point ops
//	IF T=T_LE | T=T_IE fwritef(syserr, "...in gener(): calling load(X=%n, FALSE, X | xforce(A1, T)=%n)*n", X, xforce(A1, T))
            load(X, FALSE, X=0 | xforce(A1, T))
            ENDCASE
        $)
        ENDCASE
	CASE 'M':
		emit("	@S", T)													// DC: 19-Dec-2015		- inline assembler code
		ENDCASE
	CASE 'N':
		emit("#	SECTION @S", T)
		emit("	.title @S", T)
        ENDCASE
    CASE 'T': CASE 'F':
        UNLESS T=T_LL error(6)
        X := A0
        UNLESS F0='X' & 10<=X<=15 $(
            emit("	testl %eax,%eax")
            X := 11
        $)
        code((X - 10) NEQV (F='F' -> 1, 0), A, T)
        ENDCASE
															// NB:  as of 10-May-2020 these are mostly working...
    CASE 'W':												// ...floating point ops
	CASE 'X':
        SWITCHON A INTO $(
        DEFAULT: error(8)

        CASE 1:												// (1)	A := Location (A)
            emit("	movl (,%eax,4),%eax")
            ENDCASE

        CASE 2:												// (2)	A := -A
			IF F='W' DO $(									// floating-point operation...
				emit("	pushl %eax")
				emit("	flds (%esp)")							// st[0] := a
				emit("	fchs")									// st[0] := #- a
				emit("	fstps (%esp)")
				emit("	popl %eax")								// a := #- a
				ENDCASE
			$)
            emit("	negl %eax")
            ENDCASE

        CASE 3:												// (3)	A := NOT A
			IF F='W' DO $(									// floating-point operation...
				emit("	fldz")
				emit("	fxch %st(1)")
				emit("	fucompp")
				emit("	fnstsw %ax")
				emit("	andb $69,%ah")
				emit("	xorb $64,%ah")
				emit("	jne 1f")
				emit("	movl $1065353216,%eax")
				emit("	jmp 2f")
				emit("	1:")
				emit("	movl $0,%eax")
				emit("	2:")
				ENDCASE
			$)
            emit("	xorl $-1,%eax")
            ENDCASE

        CASE 4:												// (4)	This causes a return from the current function or routine;
															//		by convention the result of a function is left in A.
															//		C := Location(P + 1)
															//		P := Location(P)
            emit("	movl 4(%ebp),%ecx")
            emit("	movl (%ebp),%ebp")
            emit("	jmp **%ecx")
            ENDCASE

        CASE 5:												// (5)	A := B * A
			IF F='W' DO $(									// floating-point operation...
			//	debug(F1, A1, T1)
				emit("	movl @A,%ecx", IA0, T0, 0, FALSE)		// collect operand1
				emit("	pushl %eax")
				emit("	flds (%esp)")							// st[0] := b
				emit("	movl %ecx,(%esp)")
				emit("	fmuls (%esp)")							// st[0] := b #* a
				emit("	fstps (%esp)")
				emit("	popl %eax")								// a := b #* a
				ENDCASE
			$)
			code(A_MUL, A0, T0)
            ENDCASE

        CASE 6: CASE 7:										// (6)	A := B / A
															// (7)	A := B REM A
			IF F='W' DO $(									// floating-point operation...
				emit("	movl @A,%ecx", IA0, T0, 0, FALSE)		// collect operand1
				emit("	pushl %eax")
				emit("	flds (%esp)")							// st[0] := b
				emit("	movl %ecx,(%esp)")
				emit("	fdivs (%esp)")							// st[0] := b #/ a
				emit("	fstps (%esp)")
				emit("	popl %eax")								// a := b #/ a
				ENDCASE
			$)
            emit("	cltd")
            code(A_DIV, A0, T0)
            TEST A=7
                emit("	movl %edx,%eax")
            ELSE
                emit("	movl %edx,91**4(%edi)")				// drop the remainder in RESULT2	(GLOBAL 91:	offset 364)
            ENDCASE

        CASE 8:												// (8)	A := B + A
			IF F='W' DO $(									// floating-point operation...
			//	debug(F1, A1, T1)
				emit("	movl @A,%ecx", IA0, T0, 0, FALSE)		// collect operand1
				emit("	pushl %eax")
				emit("	flds (%esp)")							// st[0] := b
				emit("	movl %ecx,(%esp)")
				emit("	fadds (%esp)")							// st[0] := b #+ a
				emit("	fstps (%esp)")
				emit("	popl %eax")								// a := b #+ a
				ENDCASE
			$)
            code(A_ADD, A0, T0)
            ENDCASE

        CASE 9:												// (9)	A := B - A
			IF F='W' DO $(									// floating-point operation...
				emit("	movl @A,%ecx", IA0, T_N, 0, FALSE)		// collect operand1
				emit("	pushl %eax")
				emit("	flds (%esp)")							// st[0] := b
				emit("	movl %ecx,(%esp)")
				emit("	fsubs (%esp)")							// st[0] := b #- a
				emit("	fstps (%esp)")
				emit("	popl %eax")								// a := b #- a
				ENDCASE
			$)
            code(A_SUB, A0, T0)
            ENDCASE

        CASE 10: CASE 11: CASE 12: CASE 13: CASE 14: CASE 15:	// (10)	A := B = A
															// (11)	A := B ~= A
															// (12)	A := B < A
															// (13)	A := B >= A
															// (14)	A := B > A
															// (15)	A := B <= A
			IF F='W' DO $(									// floating-point operations...
			//	debug(F1, A1, T1)
				emit("	movl @A,%ebx", IA0, T0, 0, FALSE)	// collect operand1
				SWITCHON A INTO $(			//	reverse the arguments:  'setb' and 'setbe' don't seem to work well here
					CASE 12:								// (12)	A := B < A
				//	CASE 13:								// (13)	A := B >= A
				//	CASE 14:								// (14)	A := B > A
					CASE 15:								// (15)	A := B <= A
						emit("	pushl %eax")
						emit("	flds (%esp)")					// st[0] := b
						emit("	pushl %ebx")
						emit("	flds (%esp)")					// st[0], st[1] = a, b
						ENDCASE
					DEFAULT:
						emit("	pushl %ebx")
						emit("	flds (%esp)")					// st[0] := a
						emit("	pushl %eax")
						emit("	flds (%esp)")					// st[0], st[1] = b, a
						ENDCASE
				$)
				emit("	addl $8,%esp")
				emit("	fucomip %st(1),%st")
				emit("	fstp %st")
				SWITCHON A INTO $(
					CASE 10:								// (10)	A := B = A
						emit("	seteb %al")
						ENDCASE
					CASE 11:								// (11)	A := B ~= A
						emit("	setneb %al")
						ENDCASE
					CASE 12:								// (12)	A := B < A
						emit("	seta %al")
						ENDCASE
					CASE 13:								// (13)	A := B >= A
						emit("	setae %al")
						ENDCASE
					CASE 14:								// (14)	A := B > A
						emit("	seta %al")
						ENDCASE
					CASE 15:								// (15)	A := B <= A
						emit("	setae %al")
						ENDCASE
				$)
				emit("	movzbl %al,%eax")
				emit("	negl %eax")
				ENDCASE
			$)
            code(A_CMP, A0, T0)
            IF F1='F' | F1='T'
                ENDCASE
            emit(astr(A - 10 + A_SNE))
            emit("	movzbl %al,%eax")
            emit("	decl %eax")
            ENDCASE
        CASE 16:											// (16)	A := B LSHIFT A
            TEST T0=T_N & A0<32 DO							// Intel 80386 Ref Manual: "If a shift count greater than 31 is attempted, only the bottom five bits of the shift count are used."
                code(A_SHL, A0, T0)
            ELSE $(
				emit("	movl @A,%ecx", A0, T0, 0, FALSE)
				emit("	jcxz 1f")
				emit("	2:")
				emit("	decl %ecx")
				emit("	shll $1,%eax")
				emit("	cmpl $32,%ecx")						// shift value still ge 32?
				emit("	jge 2b")
				emit("	shll %cl,%eax")
				emit("	1:")
            $)
            ENDCASE
        CASE 17:											// (17)	A := B RSHIFT A
        //	code(T0=T_N -> A_SHR, A_SR2, A0, T0)
        //	ENDCASE
            TEST T0=T_N & A0<32 DO							// Intel 80386 Ref Manual: "If a shift count greater than 31 is attempted, only the bottom five bits of the shift count are used."
                code(A_SHR, A0, T0)
            ELSE $(
				emit("	movl @A,%ecx", A0, T0, 0, FALSE)
				emit("	jcxz 1f")
				emit("	2:")
				emit("	decl %ecx")
				emit("	shrl $1,%eax")
				emit("	cmpl $32,%ecx")						// shift value still ge 32?
				emit("	jge 2b")
				emit("	shrl %cl,%eax")
				emit("	1:")
            $)
            ENDCASE
        CASE 18:											// (18)	A := B LOGAND A
        CASE 19:											// (19)	A := B LOGOR A
		/*	IF F='W' DO $(									// floating-point operation - not implemented in syntax analyser...
		 *		debug(F1, A1, T1)
		 *
		 *		emit("	fldz")
		 *		emit("	fxch %st(1)")
		 *		emit("	fucompp")
		 *		emit("	fnstsw %ax")
		 *		emit("	andb $69,%ah")
		 *		emit("	@S $64,%ah", A=18->"cmpb", "xorb")
		 *		emit("	@S", A=18->"je 2f", "jne 1f")
		 *		emit("	flds @A", IA0, T_LP, 0, FALSE)
		 *		emit("	fldz")
		 *		emit("	fxch %st(1)")
		 *		emit("	fucompp")
		 *		emit("	fnstsw %ax")
		 *		emit("	andb $69,%ah")
		 *		emit("	cmpb $64,%ah")
		 *		emit("	je 2f")
		 *		emit("	1:")
		 *		emit("	movl $1,%eax")
		 *		emit("	jmp 3f")
		 *		emit("	2:")
		 *		emit("	movl $0,%eax")
		 *		emit("	3:")
		 *		emit("	movl %eax,@A", A1, T_LP, 0, FALSE)
		 *		emit("	fildl @A", A1, T_LP, 0, FALSE)
		 *		ENDCASE
		 *	$)
		 */
            code(A=18->A_AND,A_OR, A0, T0)
            ENDCASE
        CASE 20:											// (20)	A := B NEQV A
            code(A_XOR, A0, T0)
            ENDCASE
        CASE 21:											// (21)	A := B EQV A
            emit("	xorl $-1,%eax")
            code(A_XOR, A0, T0)
            ENDCASE
        CASE 22:											// (22)	FINISH
            emit("	jmp finish")
            ENDCASE
        CASE 23:											// (23)	Switch on the value of A using data in the location
															//		addressed by C, C+1 etc.
															//		B, D := Location(C), Location(C + 1)
															//		UNTIL B = 0 DO
															//			$(	B, C := B - 1, C + 2
															//				IF A=Location(C) DO
															//					$(	D := Location(C + 1)
															//						BREAK $)       $)
															//		C := D
            emit("	movl @A,%esi", XL, T_LL, 0, FALSE)
            emit("	movl (%esi),%ecx")
            emit("	movl 4(%esi),%edx")
            emit("	jecxz 2f")
            emit("	1:")
            emit("	addl $8,%esi")
            emit("	cmpl (%esi),%eax")
            emit("	je 3f")
            emit("	loop 1b")
            emit("	2:")
            emit("	jmp **%edx")
            emit("	3:")
            emit("	jmp **4(%esi)")
            L!LN := XL
            LN := LN + 1
            XL := XL + 1
            ENDCASE
        CASE 24:											// (24)-(35)	These are implementation dependent instructions
            emit("	call selectinput")
            ENDCASE
        CASE 25:
            emit("	call selectoutput")
            ENDCASE
        CASE 26:
            emit("	call rdch")
            ENDCASE
        CASE 27:
            emit("	call wrch")
            ENDCASE
        CASE 28:
            emit("	call findinput")
            ENDCASE
        CASE 29:
            emit("	call findoutput")
            ENDCASE
        CASE 30:
            emit("	jmp stop")
            ENDCASE
        CASE 31:											// (31)	LEVEL()
            emit("	movl (%ebp),%eax")
            ENDCASE
        CASE 32:											// (32)	LONGJUMP()
            emit("	movl %ecx,%ebp")
            emit("	jmp **%eax")
            ENDCASE
        CASE 33:											// (33)	ENDREAD()
            emit("	call endread")
            ENDCASE
        CASE 34:											// (34)	ENDWRITE()
            emit("	call endwrite")
            ENDCASE
        CASE 35:											// (35)	APTOVEC()
		//	debug(F1, A1, T1)
		
		//	...preceding the APTOVEC() code we have:
		//	movl 12(%ebp),%eax
		//	movl %eax,%ebx						// Collect SIZE (in words)
		//	movl 8(%ebp),%eax					// Collect address of subroutine
		//	...then:
            emit("	movl %ebp,%esi")				// collect current frame base
            emit("	movl %eax,%ebx")
            emit("	incl %ebx")					// increment SIZE
            emit("	shll $2,%ebx")				// convert SIZE to bytes
            emit("	addl %ebx,%esi")				// calculate new frame base address
            emit("	movl (%ebp),%ebx")			// collect old frame base address in %ebx...
            emit("	movl %ebx,(%esi)")			// ...and save in new frame base location
            emit("	movl 4(%ebp),%ebx")			// collect old %eip (return address) value...
            emit("	movl %ebx,4(%esi)")			// ...and save in new frame base
            emit("	movl %ebp,%ebx")				// collect current frame base address...
            emit("	shrl $2,%ebx")				// ...convert to word address...
            emit("	movl %ebx,8(%esi)")			// ...and save in new stack frame
            emit("	movl %eax,12(%esi)")			// Save requested SIZE (words) in new stack frame
            emit("	movl %esi,%ebp")				// adjust frame base pointer for new function
            emit("	jmp **%ecx")					// Call requested subroutine!
            ENDCASE
        CASE 36:											// (36)	getbyte()
            emit("	shll $2,%eax")
            emit("	addl %eax,%ecx")
            emit("	movzbl (%ecx),%eax")
            ENDCASE
        CASE 37:											// (37)	putbyte()
		//	debug(F1, A1, T1)
			emit("#		code?")
            TEST F2='S' THEN $(
        /* args on the stack are as follows for ocode:
         *         top of stack  ->  arg1	= offset	@IA2
         *                           arg2	= vector	@IA2-1
         *                           arg3	= operand	@IA2-2
         * where the operation is  arg2%arg1 := arg3
		 */
                emit("	movl @A,%ecx", IA2-1, T_LP, 0, FALSE)			// collect BCPL vector address
                emit("	shll $2,%ecx")								// ...convert to real address
                emit("	addl @A,%ecx", IA2, T_LP, 0, FALSE)			// add byte offset
                emit("	movl @A,%eax", IA2-2, T_LP, 0, FALSE) $)		// collect operand
            ELSE TEST F2='L' THEN $(
        /* args on the stack are as follows for iclib.i code:
         *         top of stack  ->  arg1	= operand	@IA3
         *                           arg2	= offset	(%ecx)
         *                           arg3	= vector	(%eax)
         * where the operation is  arg2%arg1 := arg3
		 */
                emit("	shll $2,%eax")								// ...convert vector to real address
                emit("	addl %eax,%ecx")								// add to byte offset
                emit("	movl @A,%eax", IA3, T_LP, 0, FALSE) $)		// collect operand
			ELSE error(21)

            emit("	movb %al,(%ecx)")									// move operand to target
            ENDCASE
        CASE 38:											// (38) INPUT()
            emit("	call input")
            ENDCASE
        CASE 39:											// (39) OUTPUT()
            emit("	call output")
            ENDCASE
        CASE 40:											// (40) UNRDCH()
            emit("	call unrdch")
            ENDCASE
        CASE 41:											// (41) REWIND()
            emit("	call rewind")
			ENDCASE
        CASE 42:											// (42) GETVEC()
            emit("	call getvec")
            ENDCASE
        CASE 43:											// (43) FREEVEC()
            emit("	call freevec")
            ENDCASE
		CASE 44:											// (44)	A := ABS A
            emit("	testl %eax,%eax")
            emit("	jg 1f")									// jump if positive...
            emit("	negl %eax")								// ...otherwise negate it
            emit("	1:")
            ENDCASE
        CASE 45:											// (45)	FINDAPPEND()
			emit("	call findappend")
			ENDCASE
		CASE 46:											// (46)	Load selector - SELLD
		/* SELLD takes two argments len and sh. Its effect is equivalent to
		 * P!(S-1) := !(P!(S-1)) >> sh & mask
		 * where mask is a bit pattern containing len right justified ones. If len is zero no
		 * masking is done.
		 *
		 * args on the stack are as follows:			stack locn
         *         top of stack  ->  arg1	= shift			IA0		IA2
         *                           arg2	= length		IA3		IA4
         *                           arg3	= operand1		IA5		IA6
		 *
		 * ...so, that was easy.  Now just code it...
		 */
		CASE 47:											// (47)	Store selector - SELST
		/* SELST takes three argments op, len and sh. If op is zero, its effect is equivalent to
		 * SLCT len:sh:0 OF (P!(S-1)) := P!(S-2); S := S-2
		 * but if op is non zero it represents an assignment operator (assop) and the statement
		 * is equivalent to:
		 * SLCT len:sh:0 OF (P!(S-1)) assop:= P!(S-2); S := S-2
		 * The mapping between op and assop is given by the following table.
		 *
		 *			op assop| op assop
		 *			________|________
		 *			 1   !  |  9   +
		 *			 2  #*  | 10   -
		 *			 3  #/  | 11  <<
		 *			 4  #+  | 12  >>
		 *			 5  #-  | 13   &
		 *			 6   *  | 14   |
		 *			 7   /  | 15  EQV
		 *			 8  REM | 16  NEQV
		 *
		 * The SELST operator with len and sh both zero is used in the compilation of assop:= (+:=
		 * -:= *:= ...) assignments where the left hand side is a simple variable or a subscripted
		 * expression. For instance, the assigment v!3+:=1 might generate the following OCODE:
		 * 		LG 200 LN 1 PLUS LN 1 SELST 9 0 0
		 *
		 * args on the stack are as follows:			stack locn
         *         top of stack  ->  arg1	= shift			IA0		IA2
         *                           arg2	= length		IA3		IA4
         *                           arg3	= op			IA5		IA6
         *                           arg4	= operand1		IA7		IA8
         *                           arg5	= operand2		IA9		IA10
		 */
			$(	LET MASK=-1									// mask of all ones
			//	debug(F1, A1, T1)
				IF IA4=0 IA4:=BITSPERWORD					// (a 'length' of zero means full word mask)
				UNLESS IA4=BITSPERWORD						// if 'length' is full word cut to the chase, otherwise...
					MASK := MASK >> (BITSPERWORD-IA4)		// required 'length' mask now at rh end
				UNLESS IA2=0
					MASK := MASK << IA2						// Apply the 'shift' to the mask
						emit("#		@S code...", A=47->"SELST","SELLD")
						emit("	movl @A,%eax", A=47->IA7,IA5, T_LP, 0, FALSE)		// collect (BCPL pointer to) operand1 from stack
						emit("	movl (,%eax,4),%eax")								// collect value of operand1
				TEST A=47 THEN $(												// for SELST_..
					//	emit("	movl @A,%ecx", IA10, T10=T_R->T_N,T_LP, 0, FALSE)	// collect operand2 from stack
						emit("	movl @A,%ecx", IA9, T_LP, 0, FALSE)				// collect operand2 from stack
					SWITCHON IA6 INTO $(					// now apply the operator
					CASE  0:								// :=
					UNLESS IA2=0
						emit("	shll @A,%ecx", IA2, T_N, 0, FALSE)				// apply the 'shift'
						emit("	xorl %ecx,%eax")
					UNLESS MASK=-1
						emit("	andl @A,%eax", MASK, T_N, 0, FALSE)				// apply the prescribed MASK, unless it's all ones
						emit("	movl @A,%ecx", IA7, T_LP, 0, FALSE)				// collect (BCPL pointer to) operand1 from stack
						emit("	xorl (,%ecx,4),%eax")								// collect value of operand1
						ENDCASE
					CASE  6:								// *
						emit("	imull %ecx,%eax")						// multiply operand1 by operand2
						ENDCASE
					CASE  7:								// /
					CASE  8:								// REM
						emit("	cltd")
						emit("	idivl %ecx")							// divide operand1 by operand2
						IF IA6=8
							emit("	movl %edx,%eax")					// pass REMainder as result
						ENDCASE
					CASE  9:								// +
						emit("	addl %ecx,%eax")						// add operand2 to masked & shifted operand1
						ENDCASE
					CASE 10:								// -
						emit("	subl %ecx,%eax")						// subtract operand2 from masked & shifted operand1
						ENDCASE
					CASE 11:								// <<
						emit("	shll %cl,%eax")							// left shift operand1 by operand2	(NB: see LSHIFT above)
						ENDCASE
					CASE 12:								// >>
						emit("	shrl %cl,%eax")							// right shift operand1 by operand2	(NB: see RSHIFT above)
						ENDCASE
					CASE 13:								// &
						emit("	andl %ecx,%eax")						// logical and operand1 with operand2
						ENDCASE
					CASE 14:								// |
						emit("	orl %ecx,%eax")							// logical or operand1 with operand2
						ENDCASE
					CASE 15:								// EQV
						emit("	xorl $-1,%eax")							// bitwise equivalence of operand1 with operand2
						emit("	xorl %ecx,%eax")
						ENDCASE
					CASE 16:								// NEQV
						emit("	xorl %ecx,%eax")						// bitwise non-equivalence of operand1 with operand2
						ENDCASE
					CASE  1:								// !
																		// vecap, apparently.
					CASE  2:								// #*
					CASE  3:								// #/
					CASE  4:								// #+
					CASE  5:								// #-
					DEFAULT:
						error(22)										// Unsupported operator
						ENDCASE
					$)
						emit("	movl @A,%ecx", IA7, T_LP, 0, FALSE)		// collect (BCPL pointer to) operand1 from stack
						emit("	shll $2,%ecx")							// real pointer to operand1
						emit("	movl %eax,(%ecx)")						// save result...
				$) ELSE $(												// for SELLD...
				UNLESS MASK=-1
						emit("	andl @A,%eax", MASK, T_N, 0, FALSE)		// apply the prescribed MASK, unless it's all ones
				UNLESS IA2=0
						emit("	shrl @A,%eax", IA2, T_N, 0, FALSE)		// 'shift' result to rh end of %eax
						code(A_MV2, IA5, T_LP)							// save result...
				$)
						emit("#		...")
			$)
			ENDCASE
		CASE 48:											// (48) TIME()
			emit("	call time")
			ENDCASE
        CASE 49:											// (49)	RENAME()
			emit("	call rename")
			ENDCASE
        CASE 50:											// (50)	DELETE()
			emit("	call delete")
			ENDCASE
        CASE 51:											// (51)	GETENV()
			emit("	call getenv")
			ENDCASE
        CASE 54:											// (54)	FINDFILE()
			emit("	call findfile")
			ENDCASE
        CASE 55:											// (55)	FPOS()
			emit("	call fpos")
			ENDCASE
			
	// Floating-point operations only...
															//	NB:  Apply fixes to "...\BCPL\BCPLkit\include\syslib.b" also
		CASE 60:  $(	 									// (60) FMAKE()
															//		make an IEEE 754 32-bit single floating-point value	(code from 'syslib.b')
			LET sign, int, dec, exp, man, IEEE754 = a<0->#x80000000, 0, ?, ?, ?, 0, 0
			AND norm, ndone, p10, bit31, bitx = 31, 0, 1, #x40000000, 1
			AND a, b = ?, ?

		//	debug(F1, A1, T1)
			a := ABS IA4											// here, a	is the integer number: (if negative, sign is set to 1, above)
			b := IA2												//		 b	is the offset of the decimal point: +ve if decimal part is 0, otherwise -ve
			FOR i=1 TO ABS b DO p10 := p10*10
			TEST b>0 THEN $(
			  int := a*p10; dec := 0
			$) ELSE $(
			  int := a/p10; dec := a REM p10						// we now have integer and decimal parts as we need them...
			$)

		/*	b := ABS IA2											//		 b	is the offset of the decimal point from the rh end
		 *	FOR i=1 TO b DO p10 := p10*10
		 *	int := a/p10; dec := a REM p10							// we now have integer and decimal parts as we need them...
		 */
																			// ...let the magic begin!
//	fwritef(syserr, "*n(s)int.dec = %s%n.%n*n", sign=0->"", "-", int, dec)
							/*	____________________________________
							 *	|s|  exp   |       mantissa        |		IEEE 745 32-bit format
							 *  ^1^ 8 bits ^        23 bits        ^
							 *
							 * Convert 0.1015625 to IEEE 32-bit floating point format.
							 *	Converting:
							 *	0.1015625 	× 2 = 	0.203125 	0 	Generate 0 and continue.
							 *	0.203125 	× 2 = 	0.40625 	0 	Generate 0 and continue.
							 *	0.40625 	× 2 = 	0.8125 		0 	Generate 0 and continue.
							 *	0.8125 		× 2 = 	1.625 		1 	Generate 1 and continue.
							 *	0.625 		× 2 = 	1.25 		1 	Generate 1 and continue.
							 *	0.25 		× 2 = 	0.5 		0 	Generate 0 and continue.
							 *	0.5 		× 2 = 	1.0 		1 	Generate 1 and nothing remains.
							 *	So 0.101562510 = 0.0001101
							 *	Normalize:       0.0001101 = 1.101 × 2^-4.
							 *	Mantissa is 10100000000000000000000, exponent is -4 + 127 = 123 = 01111011, sign bit is 0. 
							 *                  __________________________________
							 *	So 0.1015625 is 0|01111011|10100000000000000000000 = x3dd00000
							 *
							 * Convert 39887.5625 to IEEE 32-bit floating point format.
							 *
							 *	The integral part is 3988710 = 10011011110011112. The fractional:
							 *	0.5625 		× 2 = 	1.125 		1 	Generate 1 and continue.
							 *	0.125 		× 2 = 	0.25 		0 	Generate 0 and continue.
							 *	0.25 		× 2 = 	0.5 		0 	Generate 0 and continue.
							 *	0.5 		× 2 = 	1.0 		1 	Generate 1 and nothing remains.
							 *	So 39887.562510 = 1001101111001111.1001
							 *	Normalize: 1001101111001111.10012 = 1.0011011110011111001 × 2^15.
							 *	Mantissa is 00110111100111110010000, exponent is 15 + 127 = 142 = 100011102, sign bit is 0. 
							 *					 __________________________________
							 *	So 39887.5625 is 0|10001110|00110111100111110010000 = x471bcf90
							 */

			IF int~=0 DO $(											// handle the integer part, moving it to the ms end of the word
			  LET bit = ?
			  $(
				bit := int&bit31
				int := int<<1										// shift left: (msb is now the sign bit!)
				norm := norm-1										// record the normalisation
				IF bit ~= 0 BREAK
				bitx := bitx<<1										// bit index to highest freed bit
			  $) REPEAT
			$)
			int := int&#x7FFFFFFF									// clear the sign bit
//	fwritef(syserr, "int=b'%32b' (x%x8), norm=%n*n", int, int, norm)

			UNLESS dec=0 $(											// no decimal part - skip processing for the mantissa
			  IF int=0 $(
				norm := 1
				bitx := #x40000000
			  $)
			  FOR i=1 TO 31 DO $(									// derive the mantissa
				LET n=?
				dec := dec<<1
				n := dec/p10
				IF n~=0 $(
			      ndone := 1										// found the first non-zero, stop calculating normalisation
			      man := man | bitx									// ...set the bit we've found (from the end of the integer)
				  dec := dec-p10
				$)
				IF int=0 & ndone=0 norm := norm+1					// putative negative normalisation if zeros so far
				bitx := bitx>>1										// bitx becomes bit(x-1), etc...
				IF dec = 0 BREAK
			  $)
			$)
//	fwritef(syserr, "man=b'%32b'*n", man)

//	fwritef(syserr, "int=%n, man=%x8, norm=%n*n", int, man, norm)
		//	TEST a=0 THEN $(
			TEST int=0 THEN $(
			  man := ((man<<norm) & #x7FFFFFFF)						// if we just have a decimal part, normalise it...
			  exp := 127-norm										// ...and adjust the exponent
			$) ELSE $(
			  exp := 127+norm
			$)

			man := (int | man)>>8
//	fwritef(syserr, "sign = %x8, exp=%x8, man=%x8*n", sign, exp&#xFF, man)
			IEEE754 := sign | ((exp&#xFF)<<23) | man				// OR the components together, and we're there!
//	fwritef(syserr, "IEEE754=x%x8 (b'%32b')*n", IEEE754, IEEE754)

			emit("	movl @A,%eax", IEEE754, T_N, 0, FALSE)
			emit("	movl %eax,@A", IA3, T_LP, 0, FALSE)
			$)
			ENDCASE

		CASE 61:											// (61) FLOAT
			emit("	pushl %eax")
			emit("	fildl (%esp)")									// st[0] := FLOAT a
			emit("	fstps (%esp)")
			emit("	popl %eax")										// a := FLOAT a
			ENDCASE

		CASE 62:											// (62) FIX
			emit("	pushl %eax")
			emit("	flds (%esp)")									// st[0] := a
			emit("	fistpl (%esp)")
			emit("	popl %eax")										// a := FIX a
			ENDCASE
        $)
											// NB: The following CASE has never been compiled - BCPL compilers don't invoke it!		DC:	22-Jun-2020
	/*	CASE xx:	$(	 									// make a BCPL decimal number pair from an IEEE 754 value
	 *		LET sign, man, exp = a&#x80000000=0->0,1 , a&#x7FFFFF, (a>>23)&#xFF		// isolate the components
	 *		AND norm=exp-127										// ...negative 'norm' means integer part will be zero
	 *		AND int, dec, bit24, bit1, p2, p10, bitpos = 0, 0, #x800000, 1, 0, 1, 24
	 *			
	 *		man := man | bit24										// re-assert the msb
	 *		TEST norm<0 THEN $(
	 *		  int := 0
	 *		  norm := ABS norm
	 *		  man := man>>norm										// un-normalise
	 *		$) ELSE $(
	 *		  LET mask=#xFFFFFF>>(23-norm)
	 *		  int := man>>norm										// un-normalise
	 *		  man := man & mask										// extract the decimal part
	 *		$)
	 *														// now convert the 'decimal' part to a base-10 number... 
	 *						/*
	 *						 * Convert IEEE 32-bit floating point value 1.101 × 2^-4 to decimal format.
	 *						 *	Mantissa is 10100000000000000000000, exponent is -4 + 127 = 123 = 01111011, sign bit is 0. 
	 *						 *	Un-normalize:       1.101 × 2^-4 = 0.0001101
	 *						 *	Converting:
	 *						 *	0			× 2^-1 =	0			Generate 0			and continue.
	 *						 *	0			× 2^-2 = 	0			Generate 0			and continue.
	 *						 *	0			× 2^-3 = 	0			Generate 0			and continue.
	 *						 *	1			× 2^-4 = 	1/16 		Generate 0.0625		and continue.
	 *						 *	1			× 2^-5 =	1/32		Generate 0.03125	and continue.
	 *						 *	0			× 2^-6 = 	0			Generate 0			and continue.
	 *						 *	1			× 2^-7 = 	1/128		Generate 0.0078125	and nothing remains.
	 *						 *											add: _________
	 *						 *												 0.1015625
	 *						 *	So 0.0001101 = 0.1015625
	 *						 */
	 *
	 *		$(
	 *		  IF man & bit1 ~=0 BREAK									// found the lsb
	 *		  bit1 := bit1<<1											// bit1 becomes bit2, etc...
	 *		  bitpos := bitpos-1
	 *		$) REPEAT
	 *
	 *		FOR i=1 TO bitpos DO p10 := p10*10							// calculate the power of 10...
	 *		p2 := 2<<bitpos												// ...and the power of 2 corresponding to the lsb
	 *		$(
	 *		  IF man&bit1~=0 DO dec := dec+(p10/p2)						// bit1 starts at the position of the lsb...
	 *		  bit1 := bit1<<1
	 *		  p2 := p2>>1
	 *		$) REPEATUNTIL p2=0
	 *					//	RESULT2 := dec								// return the decimal fraction...
	 *					//	RESULTIS int								// ...and the integer part
	 *		ENDCASE
	 *	$)
	 */

    $)
	F10, A10, T10, IA10 := F9, A9, T9, IA9
	F9, A9, T9, IA9 := F8, A8, T8, IA8
	F8, A8, T8, IA8 := F7, A7, T7, IA7
	F7, A7, T7, IA7 := F6, A6, T6, IA6
	F6, A6, T6, IA6 := F5, A5, T5, IA5
	F5, A5, T5, IA5 := F4, A4, T4, IA4
	F4, A4, T4, IA4 := F3, A3, T3, IA3
	F3, A3, T3, IA3 := F2, A2, T2, IA2
    F2, A2, T2, IA2 := F0, A0, T0, IA0
    F0, A0, T0, IA0 := F,  A,  T,  IA
    F,  A,  T,  IA  := F1, A1, T1, 0
$)

AND epilog() BE
$(  IF codename $(
      emit("	.global	code_top")
	  emit("	code_top:")
	$)
    SECT := 1
    emit("	.global G")
    emit(".align 4")
    FOR I = -neggsz TO -1 BY 1 DO $(
       emit("	.long @A	# @N", G!I, G!I -> T_LL, T_N, I, 1)
   /* TEST codename THEN SWITCHON I INTO $(
    *   CASE G_codebase:
    *     emit("	.long code_base	# @N", 0, T_LL, I)
    *     ENDCASE
    *   CASE G_codetop:
    *     emit("	.long code_top	# @N", 0, T_LL, I)
    *     ENDCASE
    *   DEFAULT:
    *     emit("	.long @A	# @N", G!I, G!I -> T_LL, T_N, I, 1)
    * $) ELSE $(
    *     emit("	.long @A	# @N", G!I, G!I -> T_LL, T_N, I, 1)
    * $)
    */
	$)
    emit("	G:")
    FOR I = 0 TO GSZ-1
        emit("	.long @A	# @N", G!I, G!I -> T_LL, T_N, I, 1)
$)

AND code(OP, A, T) BE code1(OP, A, T, 0)
AND code1(OP, A, T, R) BE emit(astr(OP), A, T, R, OP<=A_JMP)
AND data(S, A, T) BE emit(S, A, T, 0, TRUE)

AND astr(X) = VALOF
    SWITCHON X INTO $(
    CASE A_JE:  RESULTIS "	je @A"
    CASE A_JNE: RESULTIS "	jne @A"
    CASE A_JL:  RESULTIS "	jl @A"
    CASE A_JGE: RESULTIS "	jge @A"
    CASE A_JG:  RESULTIS "	jg @A"
    CASE A_JLE: RESULTIS "	jle @A"
    CASE A_JMP: RESULTIS "	jmp @A"
    CASE A_MUL: RESULTIS "	imull @A"
    CASE A_DIV: RESULTIS "	idivl @A"
    CASE A_MOV: RESULTIS "	movl @A,@R"
    CASE A_ADD: RESULTIS "	addl @A,@R"
    CASE A_SUB: RESULTIS "	subl @A,@R"
    CASE A_CMP: RESULTIS "	cmpl @A,@R"
    CASE A_SHL: RESULTIS "	shll @A,@R"
    CASE A_SHR: RESULTIS "	shrl @A,@R"
    CASE A_AND: RESULTIS "	andl @A,@R"
    CASE A_OR:  RESULTIS "	orl @A,@R"
    CASE A_XOR: RESULTIS "	xorl @A,@R"
    CASE A_MV2: RESULTIS "	movl @R,@A"
    CASE A_LEA: RESULTIS "	leal @A,@R"
	CASE A_SL2: RESULTIS "	shll %cl,@R"
	CASE A_SR2: RESULTIS "	shrl %cl,@R"
    CASE A_SNE: RESULTIS "	setne %al"
    CASE A_SE:  RESULTIS "	sete %al"
    CASE A_SGE: RESULTIS "	setge %al"
    CASE A_SL:  RESULTIS "	setl %al"
    CASE A_SLE: RESULTIS "	setle %al"
    CASE A_SG:  RESULTIS "	setg %al"

// ...and now the floating-point instructions										DC:20-November-2013

    CASE A_FML: RESULTIS ""
    CASE A_FDV: RESULTIS ""
    CASE A_FAD: RESULTIS ""
    CASE A_FSB: RESULTIS ""
    CASE A_FNE: RESULTIS ""
    CASE A_FE:	RESULTIS ""
    CASE A_FGE:	RESULTIS ""
    CASE A_FL:	RESULTIS ""
    CASE A_FLE:	RESULTIS ""
    CASE A_FG:	RESULTIS ""

    DEFAULT: error(9)
    $)

AND emit(S, A, T, X, J) BE
$(  STATIC $( PSECT=0 $)
//	IF T=T_LE|T=T_SE|T=T_SIE fwritef(syserr, "...in emit(S=*"%s*", A=%n, T=%s, X=%n, J=%n) - LN=%n*n", S, A, T=T_LE->"T_LE",T=T_SE->"T_SE","T_SIE", X, J, LN)
    UNLESS SECT=PSECT $(
        writes(SECT=0 -> ".text", ".data"); wrch('*N')
        PSECT := SECT $)
    IF LN $(
		LET ent = ?
        writes(".align 4*N")
			FOR I = 0 TO LN - 1 $(
			ent := extlist
			UNTIL ent=0 DO $(
				IF (ent!e_type=s_datalab) & (ent!e_lab=L!I) $(
					SECT := 1
					UNLESS SECT=PSECT $(
					writes(SECT=0 -> ".text", ".data"); wrch('*N')
					PSECT := SECT $)
					writef(".globl %S*n%s:*N", ent!e_name, ent!e_name)		// output the EXTERNAL name associated with this data item
//	fwritef(syserr, "...found EXTERNAL %s, L%n*n", ent!e_name, ent!e_lab)
					BREAK
				$)
				ent := ent!e_link
			$)
			wrch('L'); wrn(L!I); writes(":*N")
		$)
		LN := 0
	$)
    FOR I = 1 TO getbyte(S, 0) $(
        LET C = getbyte(S, I)
        TEST C='@' $(
            I := I + 1
            SWITCHON getbyte(S, I) INTO $(
            CASE 'A':
                TEST NOT J & (T=T_N | T=T_LL) wrch('$')
                OR IF J & T>=T_R wrch('**')
                argout(A, T)
                ENDCASE
            CASE 'N':
                argout(X, T_N)
                ENDCASE
            CASE 'S':							// Output a string					DC: 7-August-2013
                writef("%S", A)
                ENDCASE
            CASE 'R':
                argout(X, T_R)
            $)
         $) OR wrch(C)
    $)
    wrch('*N')
//	IF codecount instr := instr+1
$)

AND argout(A, T) BE
$(  TEST T=T_R | T=T_IR $(
        IF T=T_IR writes("(,")
        writes(A=0 -> "%eax", "%ecx")
        IF T=T_IR writes(",4)")
    $) OR $(
        LET K, E = T & 3, K=T_LP | K=T_LG
//	IF T=T_LE|T=T_SE|T=T_SIE fwritef(syserr, "...in argout(A=%n, T=%s)*n", A, T=T_LE->"T_LE",T=T_IE->"T_IE",T=T_SIE->"T_SIE","other")
		IF T=T_LE | T=T_IE | T=T_SIE $(
		  writes(A)								// output EXTERNAL name
		  RETURN
		$)
        IF K=T_LL | K=T_LF | T=T_LF wrch('L')
        IF E DO A := A * WORDSZ
        wrn(A)
        IF E writes(K=T_LP -> "(%ebp)", "(%edi)")
    $)
$)

AND wrn(N) BE
$(  TEST N<0 wrch('-') OR N := -N
    IF N<-9 wrn(-N / 10)
    wrch('0' - N REM 10)
$)

AND rdn() = VALOF
$(  LET X, N, I = 0, FALSE, 0
    IF ch='-' $( N := TRUE; rch() $)
    WHILE '0' <= ch <= '9' $(
        X := X * 10 - (ch - '0')
        I := I + 1
        rch()
    $)
    IF I=0 error(10)
    RESULTIS N -> X, -X
$)

AND rch() BE
$(  LET N=0
    rch1()
    UNLESS ch='/' RETURN
    $( rch1();
       N := N+1
       IF ch=endstreamch DO error(11)
       putbyte(RN, N, ch)			// Assemble any descriptive text
    $) REPEATUNTIL ch='*N'
	putbyte(RN, 0, N-1)
$) REPEAT

AND rch1() BE
$(  IF ch='*N' LINE, COL := LINE + 1, 0
    COL := COL + 1
    ch := RDCH()
$)

AND error(N, str, len) BE
$(	LET S = "?"
	selectoutput(ERR)
    UNLESS LINE=0
        writef("stdin(%N,%N): ", LINE, COL)
		
		SWITCHON N INTO $(
		DEFAULT:	S := "Unknown Error code";							ENDCASE
		CASE 1:		S := "assem() called with illegal argument";		ENDCASE
		CASE 2:		S := "Global number beyond end of GVEC array";		ENDCASE
		CASE 3:		S := "G found with modifier other than 'L'";		ENDCASE
		CASE 4:		S := "Literal value found where none expected";		ENDCASE
		CASE 5:		S := "Label number beyond end of LVEC array";		ENDCASE
		CASE 6:		S := "Unconditional jump to non-static";			ENDCASE
		CASE 7:		S := "K found with non-literal value";				ENDCASE
		CASE 8:		S := "X found with out-of-range value";				ENDCASE
		CASE 9:		S := "astr() called with illegal argument";			ENDCASE
		CASE 10:	S := "Attempt to read number finds NIL";			ENDCASE
		CASE 11:	S := "Premature end of INTCODE input file";			ENDCASE

		CASE 20:	S := "SECTION label exceeds character limit";		ENDCASE
		CASE 21:	S := "Problem identifying value for a%i:=...";		ENDCASE
		CASE 22:	S := "Unsupported operator for SELST instruction";	ENDCASE
		CASE 23:	S := "Inline assembler text exceeds length limit";	ENDCASE
		CASE 24:	S := "EXTERNAL name exceeds max length";			ENDCASE
		$)

	writef("xg error %N: %S", N, S)
	IF N=24 writef(" (%S..., %N)", str, len)
	wrch('*N')
    STOP(1)
$)

