
GET "libhdr.h"

/*	A set of debugging routines for BCPL, entry point is 'diagnose()'

	To use this code its INTCODE module should be assembled with the
	program to be debugged

 */

EXTERNAL {
	diagnose
}

MANIFEST {
	wordlim		= 32			//	bytes!
								//	offsets in codechain element:
	e_chain	= 0					//		ptr to next element
	e_addr	= 1					//		entry address of routine
	e_name	= 2					//		name of routine
}

MANIFEST {						//		Layout of the 8-bit ModR/M byte:		 2   3   3  		(SIB byte has the same structure)
	Mod		= SLCT 2:6:0		//	the Mod field of the ModR/M byte			|__|___|___|
	Reg		= SLCT 3:3:0		//	the Register field							 |   |   |
	RM		= SLCT 3:0:0		//	the RM field								 |	 |	 \___RM
}								//												 |	 \_______Reg
								//												 \___________Mod
EXTERNAL {
	code_base
	code_top
}

STATIC {
	codechain	= 0
	addr		= 0
	text		= 0
	ptxt		= 0
	modRM		= 0
	SIB			= 0
	inx			= 0
	opc			= 0
	rg1			= 0
	rg2			= 0
	ch			= 0
}

LET diagnose() BE {
  LET cis, cos = input(), output()				//	save the current input & output selections...
  LET fbase = (level()>>2)+1
  LET ptr, paddr = !fbase, ?
  LET bitsperword, msb, allones, ww = 1, 1, 1, 65
  AND pnames = TABLE 0, 0
  selectoutput(sysout)
  selectinput(sysin)							//	...and assert console input & output

  UNTIL (msb<<1)=0 DO
    bitsperword, msb, allones := bitsperword+1, msb<<1, allones<<1 | 1
  writes("*n	Entering diagnostics on a ")
  TEST (@ww)%0=65
  THEN writes("little")
  ELSE writes("big")
  writes(" ender machine*n")
  writes("	The BCPL word is ")
  writef("%n", bitsperword)
  writes(" bits long*n*n")

  ptxt := "ptxt"; FOR i=1 TO 4 DO ptxt%(i-1) := ptxt%i
  stacktrace := stacktraceII
  mapstore := mapstoreII
  mapcode()
  paddr := codechain
  UNTIL paddr=0 DO {
	UNLESS paddr!e_chain = 0 DO IF (paddr!e_chain)!e_addr<=ptr<=paddr!e_addr {
	  writef("...in %s()*n*n", (paddr!e_chain)+e_name)
	  pnames!0 := pnames!1
	  pnames!1 := (paddr!e_chain)+e_name
	  BREAK
	}
	paddr := paddr!e_chain
  }

//	writef("*n	entering diagnostic routines...*n*n")
  
  {	UNLESS ch='*n' writes("*n: ")
	ch := lcase(rdch())
	SWITCHON ch INTO {
	  CASE 'a':
		abort()

	  CASE 'd':
		codedump("name of routine to list")
		ENDCASE

	  CASE 'e':
		stop()
		ENDCASE

	  CASE 'f':
		ftable()
		ENDCASE

	  CASE 'g':
		mapstore(100)
		ENDCASE

	  CASE '?':
	  CASE 'h':
	    writes("	a	- abort*n")
	    writes("	d	- disassemble [routine_name**]*n")
	    writes("	e	- exit the program*n")
	    writes("	f	- filetable*n")
	    writes("	g	- map the GLOBAL vector <[words]>*n")
	    writes("	h or ?")
	    writes("	- this help text*n")
	    writes("	l	- list of routines*n")
		writes("	m	- view memory [decimal_word_address] <[words]>*n")
	    writes("	q	- quit diagnostics, return to calling code*n")
	    writes("	s	- stacktrace*n")
	    writes("	w	- write to memory [decimal_word_address] [hex_bytes]*n")
		wrch('*n')
		writes("		** follow with the name of a routine*n")
		writes("		  as from the 'list of routines'*n*n")
		ENDCASE

	  CASE 'l':
		listcode()
		ENDCASE

	  CASE 'm':
		dumpmem("memory address (decimal)")
		ENDCASE

	  CASE 'q':
		writef("*n	...returning to calling code*n*n")
		selectoutput(cos)
		selectinput(cis)						//	restore i/o selections
	    BREAK

	  CASE 's':
		stacktrace()
		ENDCASE

	  CASE 'w':
		writef("%s*n*n", "*n	NB: 'write to memory' is a work in progress - it may crash...")
		wrhex()
		ENDCASE

	  DEFAULT:
		UNLESS ch='*n' writef("'%c' is not implemented at this time - type 'h' for help*n", ch)
	}
  } REPEAT
}

AND lcase(ch) = 'A'<=ch<='Z' -> ch-'A'+'a', ch

AND ucase(ch) = 'a'<=ch<='z' -> ch-'a'+'A', ch

AND strcmp(s1, s2) = VALOF {
  FOR i=0 TO s1%0 DO {
	IF s1%i < s2%i RESULTIS -1		// s1 lexically less than s2
	IF s1%i > s2%i RESULTIS 1		// s1 lexically more than s2
  }
  RESULTIS 0						// s1 & s2 are identical
}

AND rdradix() = VALOF {
  LET radix, res = 10, 0

  ch := rdch() REPEATWHILE ch='*s' | ch='*t'

  SWITCHON ch INTO {
	CASE 'b': CASE 'B':
	  radix := 2
	  ch := rdch()
	  ENDCASE
	CASE 'o': CASE 'O':
	  radix := 8
	  ch := rdch()
	  ENDCASE
	CASE 'x': CASE 'X':
	  radix := 16
	  ch := rdch()
	  ENDCASE
  }
  
  { LET d = value(ch)
	IF d>=radix BREAK
	res := radix*res + d
	ch := rdch()
  } REPEATWHILE ch ~= '*n' | ch ~= '*s' | ch ~= '*t'
  RESULTIS res
}

AND wbyte(bptr, inx, val) BE {				//	a byte-address version of putbyte()...
  CODE	$[
	movl 8(%ebp),%ecx
	addl 12(%ebp),%ecx
	movl 16(%ebp),%eax
	movb %al,(%ecx)
  $]
}

AND rwhexstr(ptr) BE {
  LET octet, i = 0, 0
  {	ch := rdch() REPEATUNTIL ch ~= '*s' | ch ~= '*t'
	IF ch='*n' RETURN
	ch := value(ch)
	octet := octet | (i REM 2 -> ch,ch<<4)
	IF i REM 2 {
//	writef("ptr=%n, i/2=%n, octet=x%x2*n", ptr, i/2, octet)
//	  ptr%(i/2) := octet					//	works only for word-aligned target addresses
	  wbyte(ptr, i/2, octet)				//	works for any byte-address
	  octet := 0
	}
	i := i+1
  } REPEAT
}

AND value(ch) = '0'<=ch<='9' -> ch-'0',
                'A'<=ch<='F' -> ch-'A'+10,
                'a'<=ch<='f' -> ch-'a'+10,
                100

AND wrhex() BE {
  LET addr = ?
  ch := rdch()
  IF ch='*n' writef("%s: ", "Decimal byte address")
  addr := readn()
  rwhexstr(addr)
  ch := ' '
}

AND mapcode() BE {
  LET codeptr, entry=?, ?
  AND word = VEC wordlim/bytesperword

  UNLESS codechain=0 RETURN				// done this already
//writef("code_base=%i7,	code_top=%i7*n", (@code_base)<<2, (@code_top)<<2)
  codeptr := @code_base
  
//writef("ptxt=x%x8*n", !ptxt)
  UNTIL codeptr>=@code_top DO {
//  writef("	%i7 | x%x8*n", codeptr<<2, !codeptr)
	IF !codeptr=!ptxt {
	  LET j, k = 0, 0
	  UNTIL (codeptr+1)%j=0 DO {		//	scan the asciiz string up to the end
		word%(j+1) := (codeptr+1)%j
		j := j+1
		IF j>wordlim BREAK				//	truncate name at wordlim!
	  }
	  word%0 := j
	  k := codeptr+((j>>2)+2)			// calculate start of routine's code following |'ptxt'||'routine_name'|'\0'|
	  entry := getvec(j/bytesperword+2)
	  entry!e_addr := k<<2				// byte address of routine entry
	  FOR i=0 TO j/bytesperword DO entry!(i+e_name) := word!i
	  entry!e_chain := codechain
	  codechain := entry
	}
	codeptr := codeptr+1
  }
}

AND listcode(z) BE
{ LET codeptr, codepoint = ?, 0

  mapcode()
  codeptr := codechain
  IF codeptr=0 RETURN
  writef("*n	MAPCODE():	[%i7]*n", !(@z-1))
  UNTIL codeptr=0 DO {
    IF codepoint REM 2 = 0 wrch('*n')
	writef("*t%i7 | %s()", codeptr!e_addr, codeptr+e_name); IF (codeptr+e_name)%0<4 wrch('*t'); IF (codeptr+e_name)%0<12 wrch('*t')
	codeptr := codeptr!e_chain
	codepoint := codepoint+1
  }
  wrch('*n')
}

AND stacktraceII() BE
{ LET var1 = @var1
  LET here = level()>>2
  AND fbase = here
  AND stackhwm = stackbase
  AND pnames = VEC 1; pnames!0 := 0; pnames!1 := 0

  mapcode()
  WRITEF("*n	Stacktrace():*n*n	stackbase=%n	stackend=%n*n", STACKBASE, STACKEND)
  FOR i=STACKEND TO STACKBASE BY -1 DO {
	IF !i~=0 {
	  stackhwm := i
	  BREAK
	}
  }
  WRITEF("	stackhwm =%n*n*n", stackhwm)
  WRITES("	Stk: ...*n")
  FOR sp = stackhwm TO STACKBASE BY -1 DO {					// was...	FOR sp = var1+5 TO STACKBASE BY -1 DO
	WRITEF("	Stk: %n [%n] | x%x8 %i7	", sp, sp<<2, !sp, !sp)
	FOR j=0 TO 3 DO TEST '*s'<=sp%j<='~' THEN wrch(sp%j) OR wrch('.')
	WRITEF("  %s", sp = here ->  "<- stacktrace()", sp = STACKBASE -> "<- start()", sp = fbase -> "<-", "")
	IF sp = fbase {
//	writef("*npnames!0=%i7*n", pnames!0)
	  UNLESS (pnames!0=0) | (sp=STACKBASE) WRITEF(" %S()", pnames!0)
	  WRITEF("*n	Stk: _____________________________")
	}
	IF sp = fbase+1 {						//	this holds a pointer back into the calling code
	  LET ptr, paddr = !sp, codechain
	  UNTIL paddr=0 DO {
//	writef(paddr=%n*n", paddr)
	    UNLESS paddr!e_chain = 0 DO IF (paddr!e_chain)!e_addr<=ptr<=paddr!e_addr {
//	writef("    -> %s()", (paddr!e_chain)+e_name)
		  pnames!0 := pnames!1
		  pnames!1 := (paddr!e_chain)+e_name
//	writef("*npnames!0=%i7, pnames!1=%i7*n", pnames!0, pnames!1)
		  BREAK
		}
		paddr := paddr!e_chain
	  }
	}
	wrch('*n')
	IF fbase = sp THEN fbase := (!sp)>>2
  }
}

AND mapstoreII(ng) BE
{ GLOBAL {	GADDR	: 0
//			HIGHG	: 150
		 }
	
  LET gvec = GADDR
  AND rq, gvect = ng, ?

  ch := rdch()
  IF NOT ch='*n' rq := readn()
  ch := ' '
  ng := ng>maxglobal -> maxglobal,ng
  gvect := rq>maxglobal | rq<-neggsz -> gvec+ng, gvec+rq
//writef("rq=%n, -neggsz=%n, gvect=%n*n", rq, -neggsz, gvect)
//HIGHG := mapstoreII								// ...code test

  IF gvect=gvec+FIRSTFREEGLOBAL {					// if not user-specified, find the highest-used GLOBAL
	FOR i=999 TO -neggsz BY -1 DO IF gvec!i~=0 {
	  gvect := gvec+i
	  BREAK
	  }
	}
  WRITEF("*n	Mapstore():*n*n	GLOBAL vector starts at: %n*n*n", gvec)
  FOR gno = gvec TO gvect DO {
	TEST ((gno-gvec) REM 2) = 0 THEN WRITES("*n	") ELSE WRITES("	")

	WRITEF("G %i3 | %i7", gno-gvec, !gno)
	UNLESS codechain=0 DO {
	  LET codeptr=codechain
	  UNTIL codeptr=0 DO {
		IF !gno=codeptr!e_addr { WRITEF(" %s()%s", codeptr+e_name, (codeptr+e_name)%0<6->"*t",""); BREAK }
		codeptr := codeptr!e_chain
	  }
	  IF codeptr=0 WRITES("*t*t")
	}
  }
  WRCH('*n')
}

AND ftable() BE {
  LET ft = FILETABLE
  WRITEF("*n	File table: %n	input: %n	output: %n*n", ft, INPUT(), OUTPUT())
  FOR fp = ft TO ft+20 DO {
	WRITEF("*n		_ft %i2 | %x8", fp-ft, !fp)
	IF !fp THEN {
	  WRITEF("	*"%s*"*n", fp!21\=0 ->  fp!21, fp=ft -> "SYSIN", fp=ft+1 -> "SYSOUT", fp=ft+2 -> "SYSERR", "")
	  FOR i=0 TO 39 DO {
		TEST (i REM 16) = 0 THEN WRITES("*n			") ELSE IF (i REM 4) = 0 DO WRCH(' ')
		WRITEF("%x2", GETBYTE((!fp)>>2, i))
	  }
	  WRCH('*n')
	}
  }
  WRCH('*n')
}

AND dumpmem(prompt) BE
{ LET addr, len = ?, 1
  ch := rdch()
  IF ch='*n' writef("%s: ", prompt)
  addr := readn()
  IF terminator='*n' writef("No of words: ")
  len  := readn()
  IF len<0 len := 1
//writef("addr=%i7, len=%i3*n", addr, len)
  
  FOR i=addr+len TO addr BY -1 DO {
	writef("	Mem: %i7 |	x%x8 %i7	", i, !i, !i)
	FOR j=0 TO 3 DO { TEST '*s'<=i%j<='~' THEN wrch(i%j) OR wrch('.') }
	wrch('*n')
  }
}

AND rdtxt(prompt) = VALOF
{ LET i=0

  IF text=0 text := getvec(32)
  ch := rdch()
  IF ch='*n' writef("%s: ", prompt)
  ch := rdch() REPEATUNTIL ch~='*s' & ch~='*t' & ch~='*n'

  { i := i+1
	text%i := ch
	ch := rdch()
	IF ch='(' | ch=')' ch := rdch()
  } REPEATUNTIL ch='*s' | ch='*t' | ch='*n' | ch=')' 
  text%0 := i
  ch := '.'
  RESULTIS text
}

AND rdn(n) = VALOF {
  LET v = 0
  FOR i=n-1 TO 0 BY -1 DO {
	writef("%x2", addr%(inx+i))
	v := v<<8 | addr%(inx+i)
  }
  inx := inx+n
  RESULTIS v
}

AND reg8txt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "%al"
	CASE #x1:		RESULTIS "%cl"
	CASE #x2:		RESULTIS "%dl"
	CASE #x3:		RESULTIS "%bl"
	CASE #x4:		RESULTIS "%ah"
	CASE #x5:		RESULTIS "%ch"
	CASE #x6:		RESULTIS "%dh"
	CASE #x7:		RESULTIS "%bh"
  }
}

AND regtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "%eax"
	CASE #x1:		RESULTIS "%ecx"
	CASE #x2:		RESULTIS "%edx"
	CASE #x3:		RESULTIS "%ebx"
	CASE #x4:		RESULTIS "%esp"
	CASE #x5:		RESULTIS "%ebp"
	CASE #x6:		RESULTIS "%esi"
	CASE #x7:		RESULTIS "%edi"
  }
}

AND x2regtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS ",%eax,2"
	CASE #x1:		RESULTIS ",%ecx,2"
	CASE #x2:		RESULTIS ",%edx,2"
	CASE #x3:		RESULTIS ",%ebx,2"
	CASE #x4:		RESULTIS ""
	CASE #x5:		RESULTIS ",%ebp,2"
	CASE #x6:		RESULTIS ",%esi,2"
	CASE #x7:		RESULTIS ",%edi,2"
  }
}

AND x4regtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS ",%eax,4"
	CASE #x1:		RESULTIS ",%ecx,4"
	CASE #x2:		RESULTIS ",%edx,4"
	CASE #x3:		RESULTIS ",%ebx,4"
	CASE #x4:		RESULTIS ""
	CASE #x5:		RESULTIS ",%ebp,4"
	CASE #x6:		RESULTIS ",%esi,4"
	CASE #x7:		RESULTIS ",%edi,4"
  }
}

AND SIBregtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "%eax"
	CASE #x1:		RESULTIS "%ecx"
	CASE #x2:		RESULTIS "%edx"
	CASE #x3:		RESULTIS "%ebx"
	CASE #x4:		RESULTIS "%esp"
	CASE #x5:		RESULTIS ""				//	depending on Mod bits this may be: Mod=00 - disp32, 01 - %ebp+disp8, 10 - %ebp+disp32
	CASE #x6:		RESULTIS "%esi"
	CASE #x7:		RESULTIS "%edi"
  }
}

AND cmptxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "addl"
	CASE #x1:		RESULTIS "orl"
	CASE #x2:		RESULTIS "adcl"
	CASE #x3:		RESULTIS "sbbl"
	CASE #x4:		RESULTIS "andl"
	CASE #x5:		RESULTIS "subl"
	CASE #x6:		RESULTIS "xorl"
	CASE #x7:		RESULTIS "cmpl"
  }
}

AND shiftxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "roll"
	CASE #x1:		RESULTIS "rorl"
	CASE #x2:		RESULTIS "rcll"
	CASE #x3:		RESULTIS "rorl"
	CASE #x4:		RESULTIS "shll"
	CASE #x5:		RESULTIS "shrl"
	CASE #x6:		RESULTIS "sall"
	CASE #x7:		RESULTIS "sarl"
  }
}

AND logtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "testl"
	CASE #x1:		RESULTIS "testl"
	CASE #x2:		RESULTIS "notl"
	CASE #x3:		RESULTIS "negl"
	CASE #x4:		RESULTIS "mull"
	CASE #x5:		RESULTIS "imull"
	CASE #x6:		RESULTIS "divl"
	CASE #x7:		RESULTIS "idivl"
  }
}

AND fpxD8txt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "fadd"
	CASE #x1:		RESULTIS "fmuls"
	CASE #x2:		RESULTIS "fcom"
	CASE #x3:		RESULTIS "fcomp"
	CASE #x4:		RESULTIS "fsub"
	CASE #x5:		RESULTIS "fsubr"
	CASE #x6:		RESULTIS "fdiv"
	CASE #x7:		RESULTIS "fdivr"
  }
}

AND fpxD9txt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "flds"
	CASE #x1:		RESULTIS "fxch"
	CASE #x2:		RESULTIS "fst"
	CASE #x3:		RESULTIS "fstps"
	CASE #x4:		RESULTIS "fldenv"
	CASE #x5:		RESULTIS "fldcw"
	CASE #x6:		RESULTIS "fstenv"
	CASE #x7:		RESULTIS "fstcw"
  }
}

AND fpxDBtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "fildl"
	CASE #x1:		RESULTIS "fisttp"
	CASE #x2:		RESULTIS "fist"
	CASE #x3:		RESULTIS "fistpl"
	CASE #x4:		RESULTIS "fninit"				//	?
	CASE #x5:		RESULTIS "fld"
	CASE #x6:		RESULTIS "fcomi"
	CASE #x7:		RESULTIS "fstp"
  }
}

AND fpxDDtxt(reg) = VALOF {
  SWITCHON reg INTO {
	CASE #x0:		RESULTIS "fldl"
	CASE #x1:		RESULTIS "fisttp"
	CASE #x2:		RESULTIS "fst"
	CASE #x3:		RESULTIS "fstpl"
	CASE #x4:		RESULTIS "frstor"
	CASE #x5:		RESULTIS "fucomp"
	CASE #x6:		RESULTIS "fnsave"
	CASE #x7:		RESULTIS "fnstsw"
  }
}

AND rdregs(op, n) = VALOF {
  LET md, rg, rm = ?, ?, ?
  AND SS, ix, r32 = ?, ?, ?

  modRM := rdn(n)
  md, rg, rm := Mod::(@modRM), Reg::(@modRM), RM::(@modRM)
//	writef("*nModR/M=%x2, md=%x1, rg=%x1, rm=%x1*n", modRM, md, rg, rm)

  rg2 := op=#x88->reg8txt(rg),regtxt(rg)

  SWITCHON rm INTO {
	CASE 4:
	  TEST modRM<#xC0 THEN {			// Houston, we have a SIB...
		SIB := rdn(1)
		SS, r32, ix := Mod::(@SIB), Reg::(@SIB), RM::(@SIB)
//	writef("*nSIB=%x2, SS=%x1, r32=%x1, ix=%x1*n", SIB, SS, r32, ix)
		rg1 := SIBregtxt(rg)
		SWITCHON SS INTO {
		  CASE 0:
			rg2 := regtxt(r32)
			ENDCASE
		  CASE 1:
			rg2 := x2regtxt(r32)
			ENDCASE
		  CASE 2:
			rg2 := x4regtxt(r32)
			ENDCASE
		}
		SWITCHON op INTO {
		  CASE #xD8:					// 'f...'
			opc := fpxD8txt(rg)
			ENDCASE
		  CASE #xD9:					// 'f...'
			opc := fpxD9txt(rg)
			ENDCASE
		  CASE #xDB:					// 'f...'
			opc := fpxDBtxt(rg)
			ENDCASE
		  CASE #xDD:					// 'f...' - there are so many...
			opc := fpxDDtxt(rg)
			ENDCASE
		}
		ENDCASE
	  } ELSE GOTO noSIB
	  DEFAULT:
	  noSIB:
	  rg1 := regtxt(rm)
	  SWITCHON op INTO {
		CASE #x81:						// 'cmpl' - or something other...
		CASE #x82:
		CASE #x83:
		  opc := cmptxt(rg)
		  ENDCASE
		CASE #xC1:						// 'sh.l' - but which...
		CASE #xD1:
		  opc := shiftxt(rg)
		  ENDCASE
		CASE #xF7:						// 'testl' - etc...
		  opc := logtxt(rg)
		  ENDCASE
	  }
  }
  RESULTIS md
}

AND codedump(prompt) BE
{ LET name = rdtxt(prompt)
  AND codeptr = @code_base

//FOR i=1 TO name%0 DO name%i := ucase(name%i)			//	to upper-case or not to upper-case - that is the question...
  UNLESS codechain=0 DO {
	LET ptr = codechain
	UNTIL ptr=0 DO {
	  IF strcmp(ptr+e_name, name) = 0 { ptr := ptr!e_addr; BREAK }
	  ptr := ptr!e_chain
	}
	IF ptr=0 { writef("%s not found...*n", name); RETURN }
	codeptr := ptr>>2
	addr := codeptr
	inx := 0
	writef("*n	disassembly of %s():*n*n", name)
//	writef("addr+inx=%i7*n", addr+inx)

	UNTIL !(addr+(inx>>2)) = !ptxt DO
	{ LET op, op2, imm, lab, minus1 = ?, 0, 0, 0, -1

	  opc, rg1, rg2, SIB := "", "", "", 0
	  writef("	%i7 | %x2", (addr<<2)+inx, addr%inx)
	  op := addr%inx
	  inx := inx+1
	  SWITCHON op INTO {
		CASE #x01:
		CASE #x03:
		  opc := "addl"
		  op2 := rdregs(op, 1)
		  TEST op=#x03 {
			imm := op2=#x1->rdn(1),rdn(4)
			writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
		  } ELSE {
			writef("*t*t	%s %s,%s", opc, rg2, rg1)
		  }
		  ENDCASE
		CASE #x05:
		  opc := "addl"
		  op2 := rdregs(op, 0)
		  imm := rdn(4)
		  writef("*t	%s $%n,%s", opc, imm, rg1)
		  ENDCASE
		CASE #x0B:
		  opc := "orl"
		  op2 := rdregs(op, 1)
		  imm := rdn(1)
			writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
		  ENDCASE
		CASE #x0F:
		  opc := "movzbl"
		  op2 := rdn(1)
		  SWITCHON op2 INTO {
			CASE #x84:
			  opc := "je "
			  GOTO listx8f
			CASE #x85:
			  opc := "jne"
			  GOTO listx8f
			CASE #x8C:
			  opc := "jl "
			  GOTO listx8f
			CASE #x8D:
			  opc := "jge"
			  GOTO listx8f
			CASE #x8E:
			  opc := "jle"
			  GOTO listx8f
			CASE #x8F:
			  opc := "jg"
			listx8f:
			  imm := rdn(4)
			  writef("*t	%s  %s%n", opc, imm<0->".",".+", imm)
			  ENDCASE
			DEFAULT:
			  op2 := rdregs(op, 1)
			  writef("*t	%s (%s),%s", opc, rg1, rg2)
		  }
		  ENDCASE
		CASE #x23: 
		  opc := "andl"
		  GOTO listx3B
	/*	  op2 := rdregs(op, 1)
	 *	  imm := rdn(1)
	 *	  writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
	 *	  ENDCASE
	 */
		CASE #x25: 
		  opc := "andl"
		listx25:
		  op2 := rdregs(op, 0)
		  imm := rdn(4)
		  writef("*t	%s $%n,%s", opc, imm, rg1)
		  ENDCASE
		CASE #x2B: 
		  opc := "subl"
		  GOTO listx3B
	/*	  op2 := rdregs(op, 1)
	 *	  imm := rdn(1)
	 *	  writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
	 *	  ENDCASE
	 */
		CASE #x3B: 
		  opc := "cmpl"
		listx3B:
		  op2 := rdregs(op, 1)
		  SWITCHON op2 INTO {
			CASE 0:
			  TEST (modRM&5)=5 THEN imm := rdn(4) ELSE wrch('*t')
			  writef("*t	%s %n,%s", opc, imm, rg2)
			  ENDCASE
			CASE 1:							// one byte modifier...
			CASE 2:							// four byte modifier...
			  imm := rdn(op2=1->1,4)
			  writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
			  ENDCASE
			CASE 3:							// no modifier...
			  writef("*t	%s %s,%s", opc, rg1, rg2)
			  ENDCASE
		  }
		  ENDCASE
		CASE #x3D:
		  opc := "cmpl"
		  GOTO listx25
		CASE #x43:
		  opc := "incl"
		  rg1 := op-#x40
		  GOTO listx58
		CASE #x48:
		  opc := "decl"
		  rg1 := op-#x48
		  GOTO listx58
		CASE #x50:
		  opc := "pushl"
		  rg1 := op-#x50
		  GOTO listx58
		CASE #x58:
		  opc := "popl"
		  rg1 := op-#x58
		listx58:
		  rg1 := regtxt(rg1)
		  writef("*t*t	%s %s", opc, rg1)
		  ENDCASE
		CASE #x66:
		  imm := rdn(1)
		CASE #x90:
		  opc := ".align 4"
		  writef("*t*t	%s", opc)
		  ENDCASE
		CASE #x74:
		  opc := "je "
		  GOTO listx7f
		CASE #x75:
		  opc := "jne"
		  GOTO listx7f
		CASE #x7C:
		  opc := "jl "
		  GOTO listx7f
		CASE #x7D:
		  opc := "jge"
		  GOTO listx7f
		CASE #x7E:
		  opc := "jle"
		  GOTO listx7f
		CASE #x7F:
		  opc := "jg "
		listx7f:
		  imm := rdn(1)
		  IF imm>=#x80 THEN {
			(@minus1)%0 := imm
			imm := minus1
		  }
		  writef("*t*t	%s  %s%n", opc, imm<0->".",".+", imm)
		  ENDCASE
		CASE #x81:
		  opc := "addl"								//	'opc' is modified by 'rdregs()' if necessary... 
		  op2 := rdregs(op, 1)
		  imm := rdn(4)
		  GOTO listx83
		  ENDCASE
		CASE #x83:
		  opc := "cmpl"								//	'opc' is modified by 'rdregs()' if necessary... 
		  op2 := rdregs(op, 1)
		  imm := rdn(1)
		listx83:
		  writef("*t	%s $%n,%s", opc, imm, rg1)
		  ENDCASE
		CASE #x85:
		  opc := "testl"
		  op2 := rdregs(op, 1)
		  writef("*t*t	%s %s,%s", opc, rg1, rg1)
		  ENDCASE
		CASE #x88:
		  opc := "movb"
		  op2 := rdregs(op, 1)
		  writef("*t*t	%s %s,(%s)", opc, rg2, rg1)
		  ENDCASE
		CASE #x39: 
		CASE #x89:
		  opc := op=#x39->"cmpl","movl"
		  op2 := rdregs(op, 1)
		  TEST SIB THEN {
//	writef("*nModR/M=%x2, md=%x1, rg=%x1, rm=%x1*n", modRM, Mod::(@modRM), Reg::(@modRM), RM::(@modRM))
//	writef("*nSIB=%x2, SS=%x1, r32=%x1, ix=%x1*n", SIB, Mod::(@SIB), Reg::(@SIB), RM::(@SIB))
			IF SIB<#x40 {
			  writef("*t	%s %s,(%s)", opc, rg1, rg2)
			  GOTO xSIB
			}
			IF SIB<#x80 {
			  imm := rdn(1)
			  TEST imm=0 writef("	%s %s,(%s)", opc, rg1, rg2)
				   ELSE writef("	%s %s,%n(%s)", opc, rg1, imm, rg2)
			  GOTO xSIB
			}
			IF SIB<#xC0 {
			  imm := rdn(4)
			  TEST imm=0 writef("	%s %s,(%s)", opc, rg1, rg2)
				   ELSE writef("	%s %s,%n(%s)", opc, rg1, imm, rg2)
			}
		  xSIB:
		  } ELSE {
			SWITCHON op2 INTO {
			  CASE 0:
				IF (modRM&5)=5 imm := rdn(4)
				TEST imm=0 writef("*t	%s %s,(%s)", opc, rg2, rg1)
					 ELSE writef("*t	%s %s,%n(%s)", opc, rg2, imm, rg1)
				ENDCASE
			  CASE 1:							// one byte modifier...
			  CASE 2:							// four byte modifier...
				imm := rdn(op2=1->1,4); 		//	IF op2=1 wrch('*t')
				TEST imm=0 writef("*t	%s %s,(%s)", opc, rg2, rg1)
					 ELSE writef("*t	%s %s,%n(%s)", opc, rg2, imm, rg1)
				ENDCASE
			  CASE 3:							// no modifier...
				writef("*t*t	%s %s,%s", opc, rg2, rg1)
				ENDCASE
			}
		  }
		  ENDCASE
		CASE #x8B:
		  opc := "movl"
		  op2 := rdregs(op, 1)
		  TEST SIB THEN {
//	writef("*nModR/M=%x2, md=%x1, rg=%x1, rm=%x1*n", modRM, Mod::(@modRM), Reg::(@modRM), RM::(@modRM))
//	writef("*nSIB=%x2, SS=%x1, r32=%x1, ix=%x1*n", SIB, Mod::(@SIB), Reg::(@SIB), RM::(@SIB))
			IF SIB<#x40 {
			  writef("*t	%s (%s),%s", opc, rg2, rg1)
			  GOTO xxSIB
			}
			IF SIB<#x80 {
			  imm := rdn(1)
			  TEST imm=0 writef("*t	%s (%s),%s", opc, rg2, rg1)
				   ELSE writef("	%s %n(%s),%s", opc, imm, rg2, rg1)
			  GOTO xxSIB
			}
			IF SIB<#xC0 {
			  imm := rdn(4)
			  TEST imm=0 writef("	%s (%s),%s", opc, rg2, rg1)
				   ELSE writef("	%s %n(%s),%s", opc, imm, rg2, rg1)
			}
		  xxSIB:
		  } ELSE {
			SWITCHON op2 INTO {
			  CASE 0:						// no modifier...
				writef("*t*t	%s (%s),%s", opc, rg1, rg2)
				ENDCASE
			  CASE 1:						// one byte modifier...
				imm := rdn(1)
				TEST imm=0 writef("*t	%s (%s),%s", opc, rg1, rg2)
					 ELSE writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
				ENDCASE
			  CASE 2:						// four byte modifier...
				imm := rdn(4)
				TEST imm=0 writef("*t	%s (%s),%s", opc, rg1, rg2)
					 ELSE writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
				ENDCASE
			}
		  }
		  ENDCASE
		CASE #x8D:
		  opc := "leal"
		  op2 := rdregs(op, 1)
		  SWITCHON modRM INTO {
			CASE 5:
			CASE #xD:
			  imm := rdn(4)
			  TEST imm=0 writef("*t	%s %s", opc, rg2)
				   ELSE writef("*t	%s %n,%s", opc, imm, rg2)
			ENDCASE
			DEFAULT:
			  imm := rdn(1)
			CASE 6:
			  TEST imm=0 writef("*t	%s (%s),%s", opc, rg1, rg2)
				   ELSE writef("*t	%s %n(%s),%s", opc, imm, rg1, rg2)
		  }
		  ENDCASE
		CASE #x99:
		  opc := "cltd"
		  writef("*t*t	%s", opc)
		  ENDCASE
		CASE #xB8:
		CASE #xB9:
		CASE #xBA:
		CASE #xBB:
		CASE #xBC:
		CASE #xBD:
		CASE #xBE:						// #xB8+reg...
		  rg1 := regtxt(op-#xB8)
		CASE #xA1:
		CASE #xA3:
		  opc := "movl"
		  IF op<#xB9 THEN op2 := rdregs(op, 0)
		  imm := rdn(4)
		  TEST op=#xA3 THEN writef("*t	%s %s,%n", opc, rg1, imm)
		  ELSE writef("*t	%s $%n,%s", opc, imm, rg1)
		  ENDCASE
		CASE #xC1:
		  opc := "shrl"
		  op2 := rdregs(op, 1)
		  imm := rdn(1)
		  writef("*t	%s $%n,%s", opc, imm, rg1)
		  ENDCASE
		CASE #xC7:
		  opc := "movl"
		  op2 := rdregs(op, 1)
		  imm := rdn(1)
		  lab := rdn(4)
		  writef("*t%s $%n,%n(%s)", opc, lab, imm, rg1)
		  ENDCASE
		CASE #xD1:
		  opc := "shrl"								//	'opc' is modified by 'rdregs()' if necessary... 
		  op2 := rdregs(op, 1)
		  writef("*t*t	%s $1,%s", opc, rg1)		//	yes, the bit-shift of '1' is welded in...
		  ENDCASE
		CASE #xD8:
		  opc := "fmuls"							//	one of the many floating-point ops...
		  op2 := rdregs(op, 1)
		  writef("*t*t%s (%s)", opc, rg2)
		  ENDCASE
		CASE #xD9:
		  opc := "f..."
		  op2 := rdregs(op, 1)
		  writef("*t*t%s (%s)", opc, rg2)
		  ENDCASE
		CASE #xDB:
		  opc := "f..."
		  op2 := rdregs(op, 1)
		  writef("*t*t%s (%s)", opc, rg2)
		  ENDCASE
		CASE #xDD:
		  opc := "f..."
		  op2 := rdregs(op, 1)
		  writef("*t*t%s (%s)", opc, rg2)
		  ENDCASE
		CASE #xE2:
		  opc := "loop"
		  imm := rdn(1)
		  IF imm>=#xF0 imm := -(imm-#xF0)
		  writef("*t*t	%s .%n", opc, imm)
		  ENDCASE
		CASE #xE3:
		  opc := "jecxz"
		  imm := rdn(1)
		  writef("*t*t	%s .+%n", opc, imm)
		  ENDCASE
		CASE #xE8:
		  opc := "call"
		  imm := rdn(4)
		  writef("*t	%s .+%n", opc, imm)
		  ENDCASE
		CASE #xE9:
		CASE #xEB:
		  opc := "jmp"
		  imm := rdn(op=#xEB->1,4)
		  IF op=#xEB wrch('*t'); writef("*t	%s  .+%n", opc, imm)
		  ENDCASE
		CASE #xF7:
		  opc := "idivl"
		  op2 := rdregs(op, 1)
		  SWITCHON op2 INTO {
			CASE 0:
			  IF (modRM&5)=5 imm := rdn(4)
			  writef("*t	%s %s,%n", opc, rg1, imm)
			  ENDCASE
			CASE 1:							// one byte modifier...
			CASE 2:							// four byte modifier...
			  imm := rdn(op2=1->1,4)
			  writef("*t	%s %n(%s)", opc, imm, rg1)
			  ENDCASE
			CASE 3:							// no modifier...
			  writef("*t*t	%s %s", opc, rg1)
			  ENDCASE
		  }
		  ENDCASE
		CASE #xFF:
		  opc := "jmp"
		  op2 := rdregs(op, 1)
		  SWITCHON modRM INTO {
			CASE #x25:
			  imm := rdn(4)
			  writef("*t	%s  **%n", opc, imm)
			  ENDCASE
			DEFAULT:
			  TEST op2=1 | op2=2 THEN {
				imm := rdn(op2=1->1,4)
				writef("*t	%s  **%n(%s)", opc, imm, rg1)
			  } ELSE {
				writef("*t*t	%s  **%s", opc, rg1)
			  }
			  ENDCASE
		  }
		  ENDCASE
	  }
	  wrch('*n')
	}
  }
}


