SECTION "mial-386_w"

/*	This program compiles .mial files to .s files for 386 machines.
	Original concept and design by Martin Richards (c) August 2014

*/

/*	This version of "mial-386" has been revised to bring the code generated
	into line with code generated by Robert Nordier's 'xg' program, and
	converts MCODE to Intel-386 assembler.

	mial-386_w's resulting assembler files may be assembled and linked with
	object and library modules forming part of the BCPLkit distribution
	
															Dave Cannon
*/

GET "libhdr"



/*
The following extracts from: "The BCPL Cintsys and Cintpos User Guide"
			by Martin Richards
		
The prefixes are as follows:

	F An opcode or directive
	P A stack offset, 0 to #xFFFFFF
	G A global variable number, 0 to 65535
	K A 24-bit unsigned constant, often small in value
	W A 32-bit signed integer, used for static data and large constants
	C A byte in range 0 to 255
	L A label generated by translation phase
	M A label generated by the Sial codegenerator
	
The instructions are for an abstract machine with internal registers:

	a The main accumulator, function first arg and result register
	b The second accumulator used in dyadic operations
	c Register used by pbyt and xpbyt, and possibly corrupted by
	  some other instructions, such as mul, div, rem, xdiv and xrem
	P Pointer to the base of the current stack frame
	G Pointer to the base of the Global Vector
	PC Set by jump and call instructions

When implementing mial-386 it was necessary to decide how the Intel registers
were to be used and what the BCPL calling sequence should be. The chosen register
allocation was as follows:

	Intel register | Use							| DC: Revised usage for mial-386_w
	_______________|________________________________|______________
		%ebx	   | The A register					|	%eax
		%ecx	   | The B register					|	%ebx
		%edx	   | The C register					|	%ecx
		%eax	   | A work register				|	%edx
		%esi	   | Pointer to the GLOBAL vector	|	%edi
		%edi	   | A work register				|	%esi
		%ebp	   | The stackframe (P) pointer		|	%ebp
*/

MANIFEST {
// MIAL op codes and directives

f_lp=      1
f_lg=      2
f_ll=      3

f_llp=     4
f_llg=     5
f_lll=     6
f_lf=      7
f_lw=      8

f_l=      10
f_lm=     11

f_sp=     12
f_sg=     13
f_sl=     14

f_ap=     15
f_ag=     16
f_a=      17
f_s=      18

f_lkp=    20
f_lkg=    21
f_rv=     22
f_rvp=    23
f_rvk=    24
f_st=     25
f_stp=    26
f_stk=    27
f_stkp=   28
f_stkg=   29
f_xst=    30

f_stb=    31
f_lvind=  32
f_indb=   33
f_indb0=  34
f_indw=   35

f_k=      36
f_kpg=    37

f_neg=    38
f_not=    39
f_abs=    40

f_xdiv=   41
f_xmod=   42
f_xsub=   43

f_mul=    45
f_div=    46
f_mod=    47
f_add=    48
f_sub=    49

f_eq=     50
f_ne=     51
f_ls=     52
f_gr=     53
f_le=     54
f_ge=     55
f_eq0=    56
f_ne0=    57
f_ls0=    58
f_gr0=    59
f_le0=    60
f_ge0=    61

f_lsh=    65
f_rsh=    66
f_and=    67
f_or=     68
f_xor=    69
f_lshk=   70
f_rshk=   71

f_xch=    80
f_atb=    81
f_atc=    82
f_bta=    83
f_btc=    84
f_cta=    85
f_atblp=  86
f_atblg=  87
f_atbll=  88
f_atbl=   89

f_j=      90
f_rtn=    91

f_ikp=    93
f_ikg=    94
f_ikl=    95
f_ip=     96
f_ig=     97
f_il=     98

f_jeq=    100
f_jne=    101
f_jls=    102
f_jgr=    103
f_jle=    104
f_jge=    105
f_jeq0=   106
f_jne0=   107
f_jls0=   108
f_jgr0=   109
f_jle0=   110
f_jge0=   111

f_brk=    120
f_nop=    121
f_chgco=  122
f_mdiv=   123
f_sys=    124

f_inc1b=  130
f_inc4b=  131
f_dec1b=  132
f_dec4b=  133
f_inc1a=  134
f_inc4a=  135
f_dec1a=  136
f_dec4a=  137

f_hand =  140
f_unh  =  141
f_raise=  142

f_module=   150
f_modstart= 151
f_modend=   152
f_global=   153

f_const=    155

f_lab=      157
f_entry=    158

f_dlab=     160
f_dw=       161
f_db=       162
f_dl=       163
f_ds=       164

}

GLOBAL {
mialin:   200
asmout:   201
stdin:    202
stdout:   203

rdf:      210
rdp:      211
rdg:      212
rdk:      213
rdw:      215
rdl:      216
rdc:      218
rdcode:   219

pval:     220
gval:     221
kval:     222
wval:     224
lval:     225
mval:     226

scan:     230
cvf:      231
cvfp:     232
cvfg:     233
cvfk:     234
cvfw:     236
cvfl:     237

modname:  250
modletter:251
charv:    252
labnumber:253
}

LET start() = VALOF
{ LET argv = VEC 20
   LET v    = VEC 20
   LET cv   = VEC 256/bytesperword

   modname := v
   modname%0 := 0
   modletter := 'A'
   charv := cv
   labnumber := 0

   asmout := 0
   stdout := output()
   IF rdargs("FROM,TO/K", argv, 20)=0 DO
   { writes("Bad args for mial-386_w*n")
      RESULTIS 20
   }
   IF argv!0=0 DO argv!0 := "prog.mial"
   IF argv!1=0 DO argv!1 := "prog.s"
   mialin := findinput(argv!0)
   IF mialin=0 DO
   { writef("Trouble with file %s*n", argv!0)
      RESULTIS 20
   }
   asmout := findoutput(argv!1)
   
   IF asmout=0 DO
   { writef("Trouble with file %s*n", argv!1)
      RESULTIS 20
   }
   
   writef("Converting %s to %s*n", argv!0, argv!1)
   selectinput(mialin)
   selectoutput(asmout)

   writef("# Code generated by mial-386_w:	19-August-2015*n*n")
   writef("*n.align 4,0*n.text*n")

   scan()
   endread()
   UNLESS asmout=stdout DO endwrite()
   selectoutput(stdout)
   writef("Conversion complete*n")
   RESULTIS 0
}

AND nextlab() = VALOF
{ labnumber := labnumber+1
  RESULTIS labnumber
}

// argument may be of form Ln
AND rdcode(let) = VALOF
{ LET a, ch, neg = 0, ?, FALSE

  { ch := rdch()
    IF ch='/' DO ch := rdch() REPEATUNTIL ch='*n' | ch=endstreamch	// permit comments beginning '/'...
  } REPEATWHILE ch='*s' | ch='*n'

   IF ch=endstreamch RESULTIS -1
   UNLESS ch=let DO error("Bad item, looking for %c found %c*n", let, ch)
   ch := rdch()
   IF ch='-' DO { neg := TRUE; ch := rdch() }
   WHILE '0'<=ch<='9' DO { a := 10*a + ch - '0'; ch := rdch()  }
   RESULTIS neg -> -a, a
}

AND rdf() = rdcode('F')
AND rdp() = VALOF { pval := rdcode('P'); RESULTIS pval }
AND rdg() = VALOF { gval := rdcode('G'); RESULTIS gval }
AND rdk() = VALOF { kval := rdcode('K'); RESULTIS kval }
AND rdw() = VALOF { wval := rdcode('W'); RESULTIS wval }
AND rdl() = VALOF { lval := rdcode('L'); RESULTIS lval }
AND rdm() = VALOF { mval := rdcode('M'); RESULTIS mval }
AND rdc() = rdcode('C')

AND error(mess, a, b, c) BE
{ LET out = output()
   UNLESS out=stdout DO
   { selectoutput(stdout)
      writef(mess, a, b, c)
      selectoutput(out)
   }
   writef(mess, a, b, c)
}

AND scan() BE
{ LET op = rdf()

   SWITCHON op INTO

   { DEFAULT:       error("# Bad op %n*n", op); LOOP

      CASE -1:       RETURN
      
      CASE f_lp:     cvfp("LP") // a := P!n
                     writef("*n	movl %n(%%ebp),%%eax", 4*pval)
                     ENDCASE
      CASE f_lg:     cvfg("LG") // a := G!n
                     writef("*n	movl %n(%%edi),%%eax", 4*gval)
                     ENDCASE
      CASE f_ll:     cvfl("LL") // a := !Ln
                     writef("*n	movl L%c%n,%%eax", modletter, lval)
                     ENDCASE

      CASE f_llp:    cvfp("LLP") // a := @ P!n
                     writef("*n	leal %n(%%ebp),%%eax", 4*pval)
                     writef("*n	shrl $2,%%eax")
                     ENDCASE
      CASE f_llg:    cvfg("LLG") // a := @ G!n
                     writef("*n	leal %n(%%edi),%%eax", 4*gval)
                     writef("*n	shrl $2,%%eax")
                     ENDCASE
      CASE f_lll:    cvfl("LLL") // a := @ !Ln
                     writef("*n	movl $L%c%n,%%eax", modletter, lval)
                     writef("*n	shrl $2,%%eax")
                     ENDCASE
      CASE f_lf:     cvfl("LF") // a := byte address of Ln
                     writef("*n	leal L%c%n,%%eax", modletter, lval)
                     ENDCASE
      CASE f_lw:     cvfm("LW")
                     writef("*n	movl M%c%n,%%eax", modletter, mval)
                     ENDCASE

      CASE f_l:      cvfk("L") // a := n
                     IF kval=0 DO { writef("*n	xorl %%eax,%%eax"); ENDCASE }
                     writef("*n	movl $%n,%%eax", kval)
                     ENDCASE
      CASE f_lm:     cvfk("LM") // a := -n
                     writef("*n	movl $-%n,%%eax", kval)
                     ENDCASE
      CASE f_sp:     cvfp("SP") // P!n := a
                     writef("*n	movl %%eax,%n(%%ebp)", 4*pval)
                     ENDCASE
      CASE f_sg:     cvfg("SG") // G!n := a
                     writef("*n	movl %%eax,%n(%%edi)", 4*gval)
                     ENDCASE
      CASE f_sl:     cvfl("SL") // !Ln := a
                     writef("*n	movl %%eax,L%c%n", modletter, lval)
                     ENDCASE
      CASE f_ap:     cvfp("AP") // a := a + P!n
                     writef("*n	addl %n(%%ebp),%%eax", 4*pval)
                     ENDCASE
      CASE f_ag:     cvfg("AG") // a := a + G!n
                     writef("*n	addl %n(%%edi),%%eax", 4*gval)
                     ENDCASE
      CASE f_a:      cvfk("A") // a := a + n
                     IF kval=0 ENDCASE
                     IF kval=1  DO { writef("*n	incl %%eax"); ENDCASE }
                     IF kval=-1 DO { writef("*n	decl %%eax"); ENDCASE }
                     writef("*n	addl $%n,%%eax", kval)
                     ENDCASE
      CASE f_s:      cvfk("S")  // a := a - n
                     IF kval=0 ENDCASE
                     IF kval=1  DO { writef("*n	decl %%eax"); ENDCASE }
                     IF kval=-1 DO { writef("*n	incl %%eax"); ENDCASE }
                     writef("*n	subl $%n,%%eax", kval)
                     ENDCASE
      CASE f_lkp:    cvfkp("LKP") // a := P!n!k
                     writef("*n	movl %n(%%ebp),%%edx", 4*pval)
                     writef("*n	movl %n(,%%edx,4),%%eax", 4*kval)
                     ENDCASE
      CASE f_lkg:    cvfkg("LKG") // a := G!n!k
                     writef("*n	movl %n(%%edi),%%edx", 4*gval)
                     writef("*n	movl %n(,%%edx,4),%%eax", 4*kval)
                     ENDCASE
      CASE f_rv:     cvf("RV")  // a := ! a
                     writef("*n	movl (,%%eax,4),%%eax")
                     ENDCASE
      CASE f_rvp:    cvfp("RVP") // a := P!n!a
                     writef("*n	addl %n(%%ebp),%%eax", 4*pval)
                     writef("*n	movl (,%%eax,4),%%eax")
                     ENDCASE
      CASE f_rvk:    cvfk("RVK") // a := a!k
                     writef("*n	movl %n(,%%eax,4),%%eax", 4*kval)
                     ENDCASE
      CASE f_st:     cvf("ST") // !a := b
                     writef("*n	movl %%ebx,(%%eax)")
                     ENDCASE
      CASE f_stp:    cvfp("STP") // P!n!a := b
                     writef("*n	movl %n(%%ebp),%%edx", 4*pval)
                     writef("*n	addl %%eax,%%edx")
                     writef("*n	movl %%ebx,(,%%edx,4)")
                     ENDCASE
      CASE f_stk:    cvfk("STK") // a!n := b
                     writef("*n	movl %%ebx,%n(,%%eax,4)", 4*kval)
                     ENDCASE
      CASE f_stkp:   cvfkp("STKP")  // P!n!k := a
                     writef("*n	movl %n(%%ebp),%%edx", 4*pval)
                     writef("*n	movl %%eax,%n(,%%edx,4)", 4*kval)
                     ENDCASE
      CASE f_stkg:   cvfkg("STKG") // G!n!k := a
                     writef("*n	movl %n(%%edi),%%edx", 4*gval)
                     writef("*n	movl %%eax,%n(,%%edx,4)", 4*kval)
                     ENDCASE
      CASE f_xst:    cvf("XST") // !b := a
                     writef("*n	movl %%eax,(,%%ebx,4)")
                     ENDCASE
      CASE f_stb:    cvf("STB") // %b := a			****!!!!****
                     writef("*n	movb %%al,(,%%ebx,4)")
                     ENDCASE
      CASE f_lvind:  cvf("LVIND") // a := @ b!a
                  // writef("*n	leal (%%ebx,%%eax,4),%%eax")
                     writef("*n	addl %%ebx,%%eax")
                     writef("*n	shll $2,%%eax")
                     ENDCASE
      CASE f_indb:   cvf("INDB") // a := b % a
                /**/ writef("*n	shll $2,%%eax")
                     writef("*n	movzbl (%%eax,%%ebx),%%eax") 
                     ENDCASE
      CASE f_indb0:  cvf("INDB0") // a := % a
                /**/ writef("*n	shll $2,%%eax")
                     writef("*n	movzbl (%%eax),%%eax") 
                     ENDCASE
      CASE f_indw:   cvf("INDW") // a := b!a
                  // writef("*n	movl (%%ebx,%%eax,4),%%eax")
                     writef("*n	addl %%ebx,%%eax")
                     writef("*n	shll $2,%%eax")
                     writef("*n	movl (%%eax),%%eax")
                     ENDCASE
      CASE f_k:      cvfp("K") // Call  a(b,...) incrementing P by n
        /**/  /*     writef("*n	movl %%ebx,%n(%%ebp)", 4*(pval+2))		// pre-store arg1
               *     writef("*n	leal %n(%%ebp),%%ecx", 4*pval)
               *     writef("*n	calll **%%eax")
               */
        /**/         writef("*n	movl %%ebx,%n(%%ebp)", 4*(pval+2))		// pre-store arg1
				     writef("*n	movl %%ebp,%%ecx")
				     writef("*n	addl $%n,%%ebp", 4*pval)
				     writef("*n	movl %%ecx,(%%ebp)")
				     writef("*n	movl $1f,4(%%ebp)")
				     writef("*n	jmp **%%eax")
				     writef("*n1:")
                     ENDCASE
      CASE f_kpg:    cvfpg("KPG") // Call Gg(a,...) incrementing P by n
        /**/  /*     writef("*n	movl %%eax,%n(%%ebp)", 4*(pval+2))		// pre-store arg1 - needed for non-mial compiled libraries and OS interface
               *     writef("*n	leal %n(%%ebp),%%ecx", 4*pval)
               *     writef("*n	calll **%n(%%edi)", 4*gval)
               */
        /**/         writef("*n	movl %%eax,%n(%%ebp)", 4*(pval+2)) 		// pre-store arg1 - needed for non-mial compiled libraries and OS interface
                     writef("*n	movl %n(%%edi),%%eax", 4*gval)
				     writef("*n	movl %%ebp,%%ecx")
				     writef("*n	addl $%n,%%ebp", 4*pval)
				     writef("*n	movl %%ecx,(%%ebp)")
				     writef("*n	movl $1f,4(%%ebp)")
				     writef("*n	jmp **%%eax")
				     writef("*n1:")
                     ENDCASE
      CASE f_neg:    cvf("NEG") // a := - a
                     writef("*n	negl %%eax") 
                     ENDCASE
      CASE f_not:    cvf("NOT") // a := ~ a
                     writef("*n	notl %%eax") 
                     ENDCASE
      CASE f_abs:    cvf("ABS") // a := ABS a
                   { LET l = nextlab()
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	jge L%n", l)
                     writef("*n	negl %%eax")
                     writef("*nL%n:", l)
                     ENDCASE
                   }
      CASE f_xdiv:   cvf("XDIV") // a := a / b
                     writef("*n	cltd")
                     writef("*n	idivl %%ebx")
                     ENDCASE
      CASE f_xmod:   cvf("XMOD") // a := a MOD b
                     writef("*n	cltd")
                     writef("*n	idivl %%ebx")
                     writef("*n	movl %%edx,%%eax")
                     ENDCASE
      CASE f_xsub:   cvf("XSUB") // a := a - b
                     writef("*n	subl %%ebx,%%eax")
                     ENDCASE
      CASE f_mul:    cvf("MUL") // a := b * a; c := ?
                     writef("*n	imul %%ebx") // corrupts edx
                     ENDCASE
      CASE f_div:    cvf("DIV")  // a := b / a; c := ?
                     writef("*n	movl %%eax,%%ecx")
                     writef("*n	movl %%ebx,%%eax")
                     writef("*n	cltd")
                     writef("*n	idivl %%ecx")
                     ENDCASE
      CASE f_mod:    cvf("MOD") // a := b MOD a; c := ?
                     writef("*n	movl %%eax,%%ecx")
                     writef("*n	movl %%ebx,%%eax")
                     writef("*n	cltd")
                     writef("*n	idivl %%ecx")
                     writef("*n	movl %%edx,%%eax")
                     ENDCASE
      CASE f_add:    cvf("ADD") // a := b + a
                     writef("*n	addl %%ebx,%%eax")
                     ENDCASE
      CASE f_sub:    cvf("SUB") // a := b - a
                     writef("*n	subl %%ebx,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_eq:     cvf("EQ") // a := b = a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	seteb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_ne:     cvf("NE") // a := b ~= a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	setneb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_ls:     cvf("LS") // a := b < a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	setlb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_gr:     cvf("GR") // a := b > a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	setgb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_le:     cvf("LE") // a := b <= a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	setleb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_ge:     cvf("GE") // a := b >= a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	setgeb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_eq0:    cvf("EQ0") // a := a = 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	seteb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_ne0:    cvf("NE0") // a := a ~= 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	setneb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_ls0:    cvf("LS0") // a := a < 0
                     writef("*n	sarl $31,%%eax")
                     ENDCASE
      CASE f_gr0:    cvf("GR0") // a := a > 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	setgb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_le0:    cvf("LE0") // a := a <= 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	setleb %%al")
                     writef("*n	movzbl %%al,%%eax")
                     writef("*n	negl %%eax")
                     ENDCASE
      CASE f_ge0:    cvf("GE0") // a := a >= 0
                     writef("*n	sarl $31,%%eax")
                     writef("*n	notl %%eax")
                     ENDCASE
      CASE f_lsh:    cvf("LSH") // a := b << a; b := ?	// ebx == a; eax == b
                     writef("*n	xchgl %%eax,%%ebx")
                     writef("*n	movl %%ebx,%%ecx")
                     writef("*n	cmpl $32,%%ecx")
                     writef("*n	sbbl %%ebx,%%ebx")		// set ebx to -1 or 0 (?)
                     writef("*n	andl %%ebx,%%eax")		// set eax to b or 0
                     writef("*n	sall %%cl,%%eax")		// now shift it
                     ENDCASE
      CASE f_lshk:   cvfk("LSHK") // a := a << k
                     IF kval=0 ENDCASE
                     IF kval=1 DO { writef("*n	addl %%eax,%%eax")
                                    ENDCASE
                                  }
                     TEST 0<=kval<=31
                     THEN writef("*n	sall $%n,%%eax", kval)
                     ELSE writef("*n	xorl %%eax,%%eax")
                     ENDCASE
      CASE f_rsh:    cvf("RSH") // a := b >> a; b := ?
                     writef("*n	xchgl %%eax,%%ebx")
                     writef("*n	movl %%ebx,%%ecx")
                     writef("*n	cmpl $32,%%ecx")
                     writef("*n	sbbl %%ebx,%%ebx")		// set ebx to -1 or 0 (?)
                     writef("*n	andl %%ebx,%%eax")		// set eax to b or 0
                     writef("*n	shrl %%cl,%%eax")		// now shift it
                     ENDCASE
      CASE f_rshk:   cvfk("RSHK") // a := a >> k
                     IF kval=0 ENDCASE
                     TEST 0<=kval<=31
                     THEN writef("*n	shrl $%n,%%eax", kval)
                     ELSE writef("*n	xorl %%eax,%%eax")
                     ENDCASE
      CASE f_and:    cvf("AND") // a := b & a 
                     writef("*n	andl %%ebx,%%eax") 
                     ENDCASE
      CASE f_or:     cvf("OR") // a := b | a 
                     writef("*n	orl %%ebx,%%eax") 
                     ENDCASE
      CASE f_xor:    cvf("XOR") // a := b NEQV a
                     writef("*n	xorl %%ebx,%%eax") 
                     ENDCASE
      CASE f_xch:    cvf("XCH") // swap a and b
                     writef("*n	xchgl %%eax,%%ebx")
                     ENDCASE
      CASE f_atb:    cvf("ATB") // b := a
                     writef("*n	movl %%eax,%%ebx")
                     ENDCASE
      CASE f_atc:    cvf("ATC") // c := a
                     writef("*n	movl %%eax,%%ecx")
                     ENDCASE
      CASE f_cta:    cvf("CTA") // a := c
                     writef("*n	movl %%ecx,%%eax")
                     ENDCASE
      CASE f_bta:    cvf("BTA") // a := b
                     writef("*n	movl %%ebx,%%eax")
                     ENDCASE
      CASE f_btc:    cvf("BTC") // c := b
                     writef("*n	movl %%ebx,%%ecx")
                     ENDCASE
      CASE f_atblp:  cvfp("ATBLP") // b := a; a := P!n
                     writef("*n	movl %%eax,%%ebx")
                     writef("*n	movl %n(%%ebp),%%eax", 4*pval)
                     ENDCASE
      CASE f_atblg:  cvfg("ATBLG") // b := a; a := G!n
                     writef("*n	movl %%eax,%%ebx")
                     writef("*n	movl %n(%%edi),%%eax", 4*gval)
                     ENDCASE
      CASE f_atbl:   cvfk("ATBL") // b := a; a := k
                     writef("*n	movl %%eax,%%ebx")
                     writef("*n	movl $%n,%%eax", kval)
                     ENDCASE
      CASE f_j:      cvfl("J") // jump to Ln
                     writef("*n	jmp L%c%n", modletter, lval)
                     ENDCASE
      CASE f_rtn:    cvf("RTN") // procedure return
              /*    writef("*n	movl %%ebp,%%ecx")
               *     writef("*n	mov 4(%%ecx),%%ebp")
               *     writef("*n	jmp **(%%ecx)")
               */
                     writef("*n	movl 4(%%ebp),%%ecx")
                     writef("*n	movl (%%ebp),%%ebp")
                     writef("*n	jmp **%%ecx")
                     ENDCASE
      CASE f_ikp:    cvfkp("IKP") // a := P!n + k; P!n := a
                     writef("*n	movl %n(%%ebp),%%eax", 4*pval)
                     TEST kval=1
                     THEN writef("*n	incl %%eax")
                     ELSE TEST kval=-1
                          THEN writef("*n	decl %%eax")
                          ELSE writef("*n	addl $%n,%%eax", kval)
                     writef("*n	movl %%eax,%n(%%ebp)", 4*pval)
                     ENDCASE
      CASE f_ikg:    cvfkg("IKG") // a := G!n + k; G!n := a
                     writef("*n	movl %n(%%edi),%%eax", 4*gval)
                     TEST kval=1
                     THEN writef("*n	incl %%eax")
                     ELSE TEST kval=-1
                          THEN writef("*n	decl %%eax")
                          ELSE writef("*n	addl $%n,%%eax", kval)
                     writef("*n	movl %%eax,%n(%%edi)", 4*gval)
                     ENDCASE
      CASE f_ikl:    cvfkl("IKL") // a := !Ln + k; !Ln := a
                     writef("*n	movl L%c%n,%%eax", modletter, lval)
                     TEST kval=1
                     THEN writef("*n	incl %%eax")
                     ELSE TEST kval=-1
                          THEN writef("*n	decl %%eax")
                          ELSE writef("*n	addl $%n,%%eax", kval)
                     writef("*n	movl %%eax,L%c%n", modletter, lval)
                     ENDCASE
      CASE f_ip:     cvfp("IP") // a := P!n + a; P!n := a
                     writef("*n	addl %n(%%ebp),%%eax", 4*pval)
                     writef("*n	movl %%eax,%n(%%ebp)", 4*pval)
                     ENDCASE
      CASE f_ig:     cvfg("IG") // a := G!n + a; G!n := a
                     writef("*n	addl %n(%%edi),%%eax", 4*gval)
                     writef("*n	movl %%eax,%n(%%edi)", 4*gval)
                     ENDCASE
      CASE f_il:     cvfl("IL") // a := !Ln + a; !Ln := a
                     writef("*n	addl L%c%n,%%eax", modletter, lval)
                     writef("*n	movl %%eax,L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jeq:    cvfl("JEQ") // Jump to Ln if b = a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	je L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jne:    cvfl("JNE") // Jump to Ln if b ~= a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	jne L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jls:    cvfl("JLS") // Jump to Ln if b < a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	jl L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jgr:    cvfl("JGR") // Jump to Ln if b > a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	jg L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jle:    cvfl("JLE") // Jump to Ln if b <= a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	jle L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jge:    cvfl("JGE") // Jump to Ln if b >= a
                     writef("*n	cmpl %%eax,%%ebx")
                     writef("*n	jge L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jeq0:   cvfl("JEQ0") // Jump to Ln if a = 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	je L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jne0:   cvfl("JNE0") // Jump to Ln if a ~= 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	jne L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jls0:   cvfl("JLS0") // Jump to Ln if a < 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	jl L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jgr0:   cvfl("JGR0") // Jump to Ln if a > 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	jg L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jle0:   cvfl("JLE0") // Jump to Ln if a <= 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	jle L%c%n", modletter, lval)
                     ENDCASE
      CASE f_jge0:   cvfl("JGE0") // Jump to Ln if a >= 0
                     writef("*n	orl %%eax,%%eax")
                     writef("*n	jge L%c%n", modletter, lval)
                     ENDCASE

					 
					 
					 
					 
      // The following five opcodes are never generated by
      // the BCPL compiler:

      CASE f_brk:    cvf("BRK") // Breakpoint instruction
                     writef("*n	unimplemented")
                     ENDCASE
      CASE f_nop:    cvf("NOP") // No operation
                     ENDCASE
      CASE f_chgco:  cvf("CHGCO") // Change coroutine
                     writef("*n	unimplemented")
                     ENDCASE
      CASE f_mdiv:   cvf("MDIV") // a := Muldiv(P!3, P!4, P!5) 
                     writef("*n	unimplemented")
                     ENDCASE
      CASE f_sys:    cvf("SYS") // System function
                     writef("*n	unimplemented")
                     ENDCASE


      CASE f_module:   cvfs("MODULE") // Name of section
                       FOR i = 0 TO charv%0 DO modname%i := charv%i
                       ENDCASE
      CASE f_modstart: cvf("MODSTART") // Start of module  
                       modname%0 := 0
                       ENDCASE
      CASE f_modend:   cvf("MODEND") // End of module 
                       newline()
                       modletter := modletter+1
                       ENDCASE
      CASE f_global:   cvglobal() // Global initialisation data
                       ENDCASE
      CASE f_const:    cvconst() // Large integer constant
                       ENDCASE
      CASE f_lab:      cvfl("LAB") // Program label
                       writef("*nL%c%n:", modletter, lval)
                       ENDCASE
      CASE f_entry:    cventry() // Start of a function
                       ENDCASE
      CASE f_dlab:     cvfl("DLAB") // Static data label
                       writef("*n.data")
                       writef("*n.align 4*n")
                       writef("*nL%c%n:", modletter, lval)
                       ENDCASE
      CASE f_dw:       cvfw("DW") // Static data word
                       writef("*n	.long %n", wval)
                       ENDCASE
      CASE f_db:       cvfk("DB") // Static data byte
                       writef("*n	.byte %n", kval)
                       ENDCASE
      CASE f_dl:       cvfl("DL") // Static data word pointer
                       writef("*n	.long L%c%n", modletter, lval)
                       ENDCASE
      CASE f_ds:       cvfk("DS") // Static data space
                       FOR i = 0 TO kval DO writef("*n	.long 0")
                       ENDCASE
      CASE f_inc1b:    cvf("INC1B")   // a := ++(%a)
                    // writef("*n	movl (,%%eax,4),%%edx")
                       writef("*n	shll $2,%%eax")
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	incl %%edx")
                    // writef("*n	movl %%edx,(,%%eax,4)")
                       writef("*n	movl %%edx,(%%eax)")
                       writef("*n	movl %%edx,%%eax")
                       ENDCASE
      CASE f_inc4b:    cvf("INC4B")   // a := +++(!a)
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	addl $4,%%edx")
                       writef("*n	movl %%edx,(%%eax)")
                       writef("*n	movl %%edx,%%eax")
                       ENDCASE
      CASE f_dec1b:    cvf("DEC1B")   // a := --(%a)
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	decl %%edx")
                       writef("*n	movl %%edx,(%%eax)")
                       writef("*n	movl %%edx,%%eax")
                       ENDCASE
      CASE f_dec4b:    cvf("DEC4B")   // a := ---(!a)
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	subl $4,%%edx")
                       writef("*n	movl %%edx,(%%eax)")
                       writef("*n	movl %%edx,%%eax")
                       ENDCASE
      CASE f_inc1a:    cvf("INC1A")   // a := (%a)++
                /**/   writef("*n	shll $2,%%eax")
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	incl (%%eax)")
                       writef("*n	movl %%edx,%%eax")
                       ENDCASE
      CASE f_inc4a:    cvf("INC4A")   // a := (!a)+++
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	addl $4,%%edx")
                       writef("*n	movl %%edx,(%%eax)")
                       writef("*n	leal -4(%%edx),%%eax")
                       ENDCASE
      CASE f_dec1a:    cvf("DEC1A")   // a := (%a)--
                /**/   writef("*n	shll $2,%%eax")
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	decl (%%eax)")
                       writef("*n	movl %%edx,%%eax")
                       ENDCASE
      CASE f_dec4a:    cvf("DEC4A")   // a := (!a)---
                       writef("*n	movl (%%eax),%%edx")
                       writef("*n	subl $4,%%edx")
                       writef("*n	movl %%edx,(%%eax)")
                       writef("*n	leal 4(%%edx),%%eax")
                       ENDCASE
      CASE f_unh:      cvf("UNH")   // h := h!1
                       writef("*n	movl 20(%%edi),%%edx") // LET h = G!5
                       writef("*n	movl 4(%%edx),%%edx")  // h := h!1
                       writef("*n	movl %%edx,20(%%edi)") // G!5 := h
                       ENDCASE
      CASE f_hand:     cvfl("HAND") // a!0, a!1, a!2 := P, H, Ln
                                    // H := a
                       writef("*n	movl %%ebp,(%%eax)")   // a!0 := P
                       writef("*n	movl 20(%%edi),%%edx")
                       writef("*n	movl %%edx,4(%%eax)")  // a!1 := H
                                                         // a!2 := Ln
                       writef("*n	movl $L%c%n,8(%%eax)", modletter, lval)
                       writef("*n	movl %%eax,20(%%edi)") // H := a
                       ENDCASE
      CASE f_raise:    cvf("RAISE") // LET p,h,l := H!0, H!1, H!2
                                    // H!0, H!1, H!2 := a, b, c
                                    // P, H := p, h
                                    // GOTO l
                       writef("*n	movl 20(%%edi),%%ebp") // LET h = G!5
                       writef("*n	movl 4(%%ebp),%%edx")
                       writef("*n	movl %%edx,20(%%edi)") // G!5 := h!1
                       writef("*n	movl 8(%%ebp),%%edx")  // edx := h!2
                       writef("*n	movl %%ecx,8(%%ebp)")  // h!2 := c
                       writef("*n	movl %%ebx,4(%%ebp)")  // h!1 := b
                       writef("*n	movl (%%ebp),%%ecx")   // c   := h!0
                       writef("*n	movl %%eax,(%%ebp)")   // h!0 := a
                       writef("*n	movl %%ecx,%%ebp")     // p   := c
                       writef("*n	jmp **%%edx")          // GOTO edx
                       ENDCASE
   }

// newline()
} REPEAT

AND writeq(str, a, b, c, d, e, f, g, h, i, j, k) BE writef(str, a, b, c, d, e, f, g, h, i, j, k)

AND cvf(s)   BE writeq("*n# %s", s)
AND cvfp(s)  BE writeq("*n# %s P%n", s, rdp())
AND cvfkp(s) BE writeq("*n# %s K%n P%n", s, rdk(), rdp())
AND cvfg(s)  BE writeq("*n# %s G%n", s, rdg())
AND cvfkg(s) BE writeq("*n# %s K%n G%n", s, rdk(), rdg())
AND cvfkl(s) BE writeq("*n# %s K%n L%n", s, rdk(), rdl())
AND cvfpg(s) BE writeq("*n# %s P%n G%n", s, rdp(), rdg())
AND cvfk(s)  BE writeq("*n# %s K%n", s, rdk())
AND cvfw(s)  BE writeq("*n# %s W%n", s, rdw())
AND cvfl(s)  BE writeq("*n# %s L%n", s, rdl())
AND cvfm(s)  BE writeq("*n# %s M%n", s, rdm())

AND cvglobal() BE
{ LET n = rdk()
  LET hwm = 0

   writef("*n# GLOBAL definitions: K%n*n", n)
   modname := "G"
   IF modname%0=0 FOR i = 0 TO 4 DO modname%i := "prog"%i
// writef("*n.globl %s*n", modname)
   writef("*n.data*n")
   writef(".align 4*n")
// writef("%s:*n", modname)

   FOR i = 1 TO n DO
   { LET g = rdg()
     LET l = rdl()
     writef(".global G%n*n", g)
     writef(".equ G%n, L%c%n*n", g, modletter, l)
   }
  hwm := rdg()

}

AND rdchars() = VALOF
{ LET n = rdk()
  charv%0 := n
  FOR i = 1 TO n DO charv%i := rdc()
  RESULTIS n
}

AND cvstring() BE
{ LET lab = rdm()
   LET n = rdchars()
  // writeq("# STRING  M%n K%n", lab, n); FOR i = 1 TO n DO writef(" C%n", charv%i)
   writef("*n.data")
   writef("*n	.align 4")
   writef("*nM%c%n:", modletter, lab)
   FOR i = 0 TO n DO writef("*n	.byte %n", charv%i)
   writef("*n	.text")
}

AND cvconst() BE
{ LET lab = rdm()
   LET w = rdw()
// writef("# CONST   M%n W%n", lab, w)
   writef("*n.data")
   writef("*n	.align 4")
   writef("*nM%c%n:", modletter, lab)
   writef("*n	.long %n", w)
   writef("*n	.text")
}

AND cvstatic() BE
{ LET lab = rdl()
   LET n = rdk()
   writef("# STATIC  L%n K%n", lab, n)
   writef("*n.data")
   writef("*n	.align 4")
   writef("*nL%c%n:", modletter, lab)
   FOR i = 1 TO n DO { writef("*n# W%n", rdw())
                       writef("*n	.long %n", wval)
                     }
   writef("*n	.text")
}

AND cvfs(s) BE
{ LET n = rdchars()
// writef("# %s K%n", s, n)
// FOR i = 1 TO n DO writef(" C%n", charv%i)
   writef("*n# %s ", s)
   FOR i = 1 TO n DO wrch(charv%i)
}

AND cventry() BE
{ LET n = rdchars()
   LET op = rdf()
   LET lab = rdl()
   writef("*n*n# Entry to: %s", charv)
  // writeq("# %s K%n", "ENTRY", n); FOR i = 1 TO n DO writef(" C%n", charv%i)
  // newline()
  writef("*n.align 4,0*n.text")
  TEST op=f_lab THEN writef("*n# LAB L%n", lab)
                ELSE writef("*n# Bad op F%n L%n*n", op, lab)
	writef("*nL%c%n:", modletter, lab)
/*	writef("*n	pop (%%ecx)")
 *	writef("*n	mov %%ebp,4(%%ecx)")	// P!0 := previous stackbase
 *	writef("*n	mov %%ecx,%%ebp")
 *	writef("*n	mov 12(%%ebp),%%eax")	// P!3 := arg2
 */
}
