// $Id: syslib.b,v 1.1 2018/08/07 19:53:28 rn Exp $

// SECTION "SYSLIB"

/*
 *	This module is a minimalist implementation of the sys() function
 *	required by later versions of the BCPL compiler...
 *	
 *	It is most easily included in those programs by the addition of the
 *	line:
 *		GET "syslib.b"
 *	near the head of the file.						Dave C	9-Sep-2014
 */

//	GET "libhdr"				// should not be necessary - the code needing sys(...) should...
//	GET "bcplfecg_x"			// ...already have GOT these

MANIFEST $(
	unsupported	= #Xffffffff
$)


LET sys(fn, op, a, b, c, d, e, f, g) = VALOF
$(  LET bw = VEC 9
// fwritef(syserr, "*n...sys():  fn=%n, op=%n, a=%n, b=%n*n", fn, op, a, b)
	SWITCHON fn INTO $(
		DEFAULT:	RESULTIS unsupported
		
		CASE Sys_flt:										// NB:  Apply fixes to "...\BCPL\BCPLkit\BCPL 2020-05-13\codegens\xg_i.b" also
		  SWITCHON op INTO $(
			DEFAULT:	RESULTIS unsupported

			CASE fl_mk:	$(	 												// make an IEEE 754 32-bit single floating-point value
				LET sign, int, dec, exp, man, IEEE754 = a<0->#x80000000, 0, ?, ?, ?, 0, 0
				AND norm, ndone, p10, bit31, bitx = 31, 0, 1, #x40000000, 1
				a := ABS a													// here, a	is the integer number: if negative, set sign to 1
				b := b														//		 b	is the offset of the decimal point: +ve if decimal part is 0, otherwise -ve
				FOR i=1 TO ABS b DO p10 := p10*10
				TEST b>0 THEN $(
				  int := a*p10; dec := 0
				$) ELSE $(
				  int := a/p10; dec := a REM p10							// we now have integer and decimal parts as we need them...
				$)

			/*	b := ABS b													//		 b	is the offset of the decimal point from the rh end
			 *	FOR i=1 TO b DO p10 := p10*10
			 *	int := a/p10; dec := a REM p10								// we now have integer and decimal parts as we need them...
			 */
			 																// ...let the magic begin!
// fwritef(syserr, "*n(s)int.dec = %s%n.%n*n", sign=0->"", "-", int, dec)
							/*	____________________________________
							 *	|s|  exp   |       mantissa        |		IEEE 745 32-bit format
							 *  ^1^ 8 bits ^        23 bits        ^
							 *
							 * Convert 0.1015625 to IEEE 32-bit floating point format.
							 *	Converting:
							 *	0.1015625 	× 2 = 	0.203125 	0 	Generate 0 and continue.
							 *	0.203125 	× 2 = 	0.40625 	0 	Generate 0 and continue.
							 *	0.40625 	× 2 = 	0.8125 		0 	Generate 0 and continue.
							 *	0.8125 		× 2 = 	1.625 		1 	Generate 1 and continue.
							 *	0.625 		× 2 = 	1.25 		1 	Generate 1 and continue.
							 *	0.25 		× 2 = 	0.5 		0 	Generate 0 and continue.
							 *	0.5 		× 2 = 	1.0 		1 	Generate 1 and nothing remains.
							 *	So 0.101562510 = 0.0001101
							 *	Normalize:       0.0001101 = 1.101 × 2^-4.
							 *	Mantissa is 10100000000000000000000, exponent is -4 + 127 = 123 = 01111011, sign bit is 0. 
							 *                  __________________________________
							 *	So 0.1015625 is 0|01111011|10100000000000000000000 = x3dd00000
							 *
							 * Convert 39887.5625 to IEEE 32-bit floating point format.
							 *
							 *	The integral part is 3988710 = 10011011110011112. The fractional:
							 *	0.5625 		× 2 = 	1.125 		1 	Generate 1 and continue.
							 *	0.125 		× 2 = 	0.25 		0 	Generate 0 and continue.
							 *	0.25 		× 2 = 	0.5 		0 	Generate 0 and continue.
							 *	0.5 		× 2 = 	1.0 		1 	Generate 1 and nothing remains.
							 *	So 39887.562510 = 1001101111001111.1001
							 *	Normalize: 1001101111001111.10012 = 1.0011011110011111001 × 2^15.
							 *	Mantissa is 00110111100111110010000, exponent is 15 + 127 = 142 = 100011102, sign bit is 0. 
							 *					 __________________________________
							 *	So 39887.5625 is 0|10001110|00110111100111110010000 = x471bcf90
							 */

				IF int~=0 DO $(												// handle the integer part, moving it to the ms end of the word
				  LET bit = ?
				  $(
					bit := int&bit31
					int := int<<1											// shift left: (msb is now the sign bit!)
					norm := norm-1											// record the normalisation
					IF bit ~= 0 BREAK
					bitx := bitx<<1											// bit index to highest freed bit
				  $) REPEAT
				$)
				int := int&#x7FFFFFFF										// clear the sign bit
// fwritef(syserr, "int=b'%s' (x%x8)*n", binaryw(int, bw), int)

				UNLESS dec=0 $(											// no decimal part - skip processing for the mantissa
				  IF int=0 $(
					norm := 1
					bitx := #x40000000
				  $)
				  FOR i=1 TO 31 DO $(										// derive the mantissa
					LET n=?
					dec := dec<<1
					n := dec/p10
					IF n~=0 $(
				      ndone := 1											// found the first non-zero, stop calculating normalisation
				      man := man | bitx										// ...set the bit we've found (from the end of the integer)
					  dec := dec-p10
					$)
					IF int=0 & ndone=0 norm := norm+1						// putative negative normalisation if zeros so far
					bitx := bitx>>1											// bitx becomes bit(x-1), etc...
					IF dec = 0 BREAK
				  $)
				$)
// fwritef(syserr, "man=b'%s'*n", binaryw(man, bw))

// fwritef(syserr, "int=%n, man=%x8, norm=%n*n", int, man, norm)
			//	TEST a=0 THEN $(
				TEST int=0 THEN $(
				  man := ((man<<norm) & #x7FFFFFFF)							// if we just have a decimal part, normalise it...
				  exp := 127-norm											// ...and adjust the exponent
				$) ELSE $(
				  exp := 127+norm
				$)

				man := (int | man)>>8
// fwritef(syserr, "sign = %x8, exp=%x8, man=%x8*n", sign, exp&#xFF, man)
				 IEEE754 := sign | ((exp&#xFF)<<23) | man					// OR the components together, and we're there!
// fwritef(syserr, "IEEE754=x%x8 (b'%s')*n", IEEE754, binaryw(IEEE754, bw))
							RESULTIS IEEE754
			  $)

			CASE fl_unmk:	$(	 											// make a BCPL decimal number pair from an IEEE 754 value
				LET sign, man, exp = a&#x80000000=0->0,1 , a&#x7FFFFF, (a>>23)&#xFF		// isolate the components
				AND norm=exp-127											// ...negative 'norm' means integer part will be zero
				AND int, dec, bit24, bit1, p2, p10, bitpos = 0, 0, #x800000, 1, 0, 1, 24
				
				man := man | bit24											// re-assert the msb
				TEST norm<0 THEN $(
				  int := 0
				  norm := ABS norm
				  man := man>>norm											// un-normalise
				$) ELSE $(
				  LET mask=#xFFFFFF>>(23-norm)
				  int := man>>norm											// un-normalise
				  man := man & mask											// extract the decimal part
				$)
															// now convert the 'decimal' part to a base-10 number... 
							/*
							 * Convert IEEE 32-bit floating point value 1.101 × 2^-4 to decimal format.
							 *	Mantissa is 10100000000000000000000, exponent is -4 + 127 = 123 = 01111011, sign bit is 0. 
							 *	Un-normalize:       1.101 × 2^-4 = 0.0001101
							 *	Converting:
							 *	0			× 2^-1 =	0			Generate 0			and continue.
							 *	0			× 2^-2 = 	0			Generate 0			and continue.
							 *	0			× 2^-3 = 	0			Generate 0			and continue.
							 *	1			× 2^-4 = 	1/16 		Generate 0.0625		and continue.
							 *	1			× 2^-5 =	1/32		Generate 0.03125	and continue.
							 *	0			× 2^-6 = 	0			Generate 0			and continue.
							 *	1			× 2^-7 = 	1/128		Generate 0.0078125	and nothing remains.
							 *											add: _________
							 *												 0.1015625
							 *	So 0.0001101 = 0.1015625
							 */

				$(
				  IF man & bit1 ~=0 BREAK									// found the lsb
				  bit1 := bit1<<1											// bit1 becomes bit2, etc...
				  bitpos := bitpos-1
				$) REPEAT

				FOR i=1 TO bitpos DO p10 := p10*10							// calculate the power of 10...
				p2 := 2<<bitpos												// ...and the power of 2 corresponding to the lsb
				$(
				  IF man&bit1~=0 DO dec := dec+(p10/p2)						// bit1 starts at the position of the lsb...
				  bit1 := bit1<<1
				  p2 := p2>>1
				$) REPEATUNTIL p2=0
							RESULT2 := dec									// return the decimal fraction...
							RESULTIS int									// ...and the integer part
			  $)

			CASE fl_float:	RESULTIS FLOAT a
			CASE fl_fix:	RESULTIS FIX   a
			CASE fl_abs:	RESULTIS #ABS  a
			CASE fl_mul:	RESULTIS a  #* b
			CASE fl_div:	RESULTIS a  #/ b
			CASE fl_add:	RESULTIS a  #+ b
			CASE fl_sub:	RESULTIS a  #- b
			CASE fl_neg:	RESULTIS    #- a
			CASE fl_eq:		RESULTIS a  #= b
			CASE fl_ne:		RESULTIS a #~= b
			CASE fl_ls:		RESULTIS a  #< b
			CASE fl_gr:		RESULTIS a  #> b
			CASE fl_le:		RESULTIS a #<= b
			CASE fl_ge:		RESULTIS a #>= b
		$)
	$)
$)

AND fwritef(strm, fmt, a, b, c, d, e, f, g, h, i, j, k) BE
$( LET cos=output()
   selectoutput(strm)
   writef(fmt, a, b, c, d, e, f, g, h, i, j, k)
   selectoutput(cos)
$)

AND binaryw(w, v) = VALOF
$( LET bit31=#x80000000
   AND i, j = 32, 1
   $( LET bit=w&bit31
// fwritef(syserr, "i=%n, j=%n, bit31=x%x8*n", i, j, bit31)
      v%j := bit ~= 0 -> '1', '0'
	  bit31 := bit31>>1						// bit31 becomes bit30, etc...
	  j := j+1
	  i := i-1
   $) REPEATUNTIL i=0
   v%0 := 32
   RESULTIS v
$)